# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

""" Creating anomaly detector

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_anomaly_steps as anomaly_create
from . import create_multimodel_steps as mm_create

class TestAnomaly:
    """Test anomaly detector methods"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating an anomaly detector from a dataset and a dataset list:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            Then I create an anomaly detector from a dataset
            And I wait until the anomaly detector is ready less than <model_wait> secs
            And I check the anomaly detector stems from the original dataset
            And I store the dataset id in a list
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I store the dataset id in a list
            Then I create an anomaly detector from a dataset list
            And I wait until the anomaly detector is ready less than 'model_wait'> secs
            And I check the anomaly detector stems from the original dataset list
        """
        show_doc(self.test_scenario1)
        headers = ["data", "source_wait", "dataset_wait", "model_wait"]
        examples = [
            ['data/tiny_kdd.csv', '40', '40', '100']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(
                self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            anomaly_create.i_create_an_anomaly_from_dataset(
                self, shared=example["data"])
            anomaly_create.the_anomaly_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            anomaly_create.i_check_anomaly_dataset_and_datasets_ids(self)
            mm_create.i_store_dataset_id(self)
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            mm_create.i_store_dataset_id(self)
            anomaly_create.i_create_an_anomaly_from_dataset_list(self)
            anomaly_create.the_anomaly_is_finished_in_less_than(
                self, example["model_wait"])
            anomaly_create.i_check_anomaly_datasets_and_datasets_ids(self)

    def test_scenario2(self):
        """
        Scenario: Successfully creating an anomaly detector from a dataset and generating the anomalous dataset:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <time_1> secs
            And I create a dataset
            And I wait until the dataset is ready less than <time_2> secs
            Then I create an anomaly detector of <rows> anomalies from a dataset
            And I wait until the anomaly detector is ready less than <time_4> secs
            And I create a dataset with only the anomalies
            And I wait until the dataset is ready less than <time_3> secs
            And I check that the dataset has <rows> rows

            Examples:
            | data                       | time_1  | time_2 | time_3 |time_4|  rows|
            | ../data/iris_anomalous.csv | 40      | 40     | 80     | 40   |  1
        """
        show_doc(self.test_scenario2)
        headers = ["data", "source_wait", "dataset_wait", "model_wait", "rows"]
        examples = [
            ['data/iris_anomalous.csv', '40', '40', '80', '1']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            anomaly_create.i_create_an_anomaly_with_top_n_from_dataset(
                self, example["rows"])
            anomaly_create.the_anomaly_is_finished_in_less_than(
                self, example["model_wait"])
            anomaly_create.create_dataset_with_anomalies(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["model_wait"])
            anomaly_create.the_dataset_has_n_rows(self, example["rows"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2018-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from bigml.api import HTTP_CREATED
from bigml.api import FINISHED, FAULTY
from bigml.io import UnicodeReader

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_, ok_


def i_create_a_batch_projection(step):
    """Step: I create a batch projection for the dataset with the PCA"""
    dataset = world.dataset.get('resource')
    pca = world.pca.get('resource')
    resource = world.api.create_batch_projection(pca, dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.batch_projection = resource['object']
    world.batch_projections.append(resource['resource'])


def wait_until_batch_projection_status_code_is(step, code1, code2, secs):
    """Step: I wait until the batch projection status code is either <code1>
    or <code2> less than <secs>
    """
    world.batch_projection = wait_until_status_code_is(
        code1, code2, secs, world.batch_projection)


def the_batch_projection_is_finished_in_less_than(step, secs):
    """Step: I wait until the batch projection is ready less than <secs>"""
    wait_until_batch_projection_status_code_is(step, FINISHED, FAULTY, secs)


def i_download_projections_file(step, filename):
    """Step: I download the created projections file to <filename>"""
    file_object = world.api.download_batch_projection(
        world.batch_projection, filename=res_filename(filename))
    ok_(file_object is not None)
    world.output = file_object


def i_check_projections(step, check_file):
    """Step: the batch projection file is like <check_file>"""
    with UnicodeReader(world.output) as projection_rows:
        with UnicodeReader(res_filename(check_file)) as test_rows:
            check_csv_rows(projection_rows, test_rows)


def check_csv_rows(projections, expected):
    """Checking expected projections"""
    for projection in projections:
        eq_(projection, next(expected))



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2014-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY
from bigml.association import Association

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_


def i_check_association_name(step, name):
    """Step: the association name is <name>"""
    association_name = world.association['name']
    eq_(name, association_name)


def i_create_an_association_from_dataset(step, shared=None):
    """Step: I create an association from a dataset"""
    if shared is None or world.shared.get("association", {}).get("shared") is None:
        dataset = world.dataset.get('resource')
        resource = world.api.create_association(dataset, {'name': 'new association'})
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.association = resource['object']
        world.associations.append(resource['resource'])


def i_create_an_association_from_dataset_with_params(step, parms=None):
    """Step: I create an association from a dataset with params <parms>"""
    dataset = world.dataset.get('resource')
    if parms is not None:
        parms = json.loads(parms)
    else:
        parms = {}
    parms.update({'name': 'new association'})
    resource = world.api.create_association(dataset, parms)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.association = resource['object']
    world.associations.append(resource['resource'])


def i_create_an_association_with_strategy_from_dataset(step, strategy):
    """Step: I create an association with search strategy <strategy>
    from a dataset
    """
    dataset = world.dataset.get('resource')
    resource = world.api.create_association(
        dataset, {'name': 'new association', 'search_strategy': strategy})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.association = resource['object']
    world.associations.append(resource['resource'])


def i_update_association_name(step, name):
    """Step: I update the association name to <name>"""
    resource = world.api.update_association(world.association['resource'],
                                            {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.association = resource['object']


def wait_until_association_status_code_is(step, code1, code2, secs):
    """Step: I wait until the association status code is either <code1> or
    <code2> less than <secs>
    """
    world.association = wait_until_status_code_is(
        code1, code2, secs, world.association)


def the_association_is_finished_in_less_than(step, secs, shared=None):
    """Steps: I wait until the association is ready less than <secs>"""
    if shared is None or world.shared.get("association", {}).get(shared) is None:
        wait_until_association_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if "association" not in world.shared:
                world.shared["association"] = {}
            world.shared["association"][shared] = world.association
    else:
        world.association = world.shared["association"][shared]
        print("Reusing %s" % world.association["resource"])


def i_create_a_local_association(step):
    """Step: I create a local association"""
    step.bigml["local_association"] = Association(world.association)


def i_get_rules_for_item_list(step, item_list):
    """Step: I get the rules for <item_list>"""
    world.association_rules = step.bigml["local_association"].get_rules(
        item_list=item_list)


def the_first_rule_is(step, rule):
    """Step: the first rule is <rule>"""
    found_rules = []
    for a_rule in world.association_rules:
        found_rules.append(a_rule.to_json())
    eq_(rule, found_rules[0])


def i_export_association(step, filename):
    """Step: I export the association"""
    world.api.export(world.association.get('resource'),
                     filename=res_filename(filename))


def i_create_local_association_from_file(step, export_file):
    """Step: I create a local association from file <export_file>"""
    step.bigml["local_association"] = Association(res_filename(export_file))


def check_association_id_local_id(step):
    """Step: the association ID and the local association ID match"""
    eq_(step.bigml["local_association"].resource_id,
        world.association["resource"])


def clone_association(step, association):
    """Step: I clone association"""
    resource = world.api.clone_association(association,
                                           {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.association = resource['object']
    # save reference
    world.associations.append(resource['resource'])


def the_cloned_association_is(step, association):
    """The association is a clone"""
    eq_(world.association["origin"], association)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import,invalid-name
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Checking webhooks secrets

"""
import json

from collections import OrderedDict
from bigml.webhooks import check_signature

from .world import world, setup_module, teardown_module, show_doc, \
    show_method, ok_


BIGML_SECRET = 'mysecret'

BIGML_REQUEST_MOCKUP = {
    "body": {
        'event': 'finished',
        'message': 'The source has been created',
        'resource': 'source/627eceb1d432eb7338001d4b',
        'timestamp': '2022-05-13 21:33:39 GMT'
    },
    "META": {
        'HTTP_X_BIGML_SIGNATURE': "sha1=af38d979e8582d678653a8059ca0821daeedebbd"
    }
}


class RequestMockup:
    """Test for webhooks with secrets"""

    def __init__(self, request_dict):
        self.body = json.dumps(request_dict["body"], sort_keys=True)
        self.meta = request_dict["META"]


class TestWebhook:
    """Testing webhooks"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Testing webhook secret signature
        """
        show_doc(self.test_scenario1)
        ok_(check_signature(RequestMockup(BIGML_REQUEST_MOCKUP),
                            BIGML_SECRET))



# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Class to store Dataset transformations based on the Dataset API response

"""
import os
import logging

from bigml.fields import Fields, sorted_headers, get_new_fields
from bigml.api import get_api_connection, get_dataset_id, get_status
from bigml.basemodel import get_resource_dict
from bigml.util import DEFAULT_LOCALE, use_cache, cast, load, dump, dumps
from bigml.constants import FINISHED
from bigml.flatline import Flatline
from bigml.featurizer import Featurizer

# avoiding tensorflow info logging
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
logging.getLogger('tensorflow').setLevel(logging.ERROR)

#pylint: disable=locally-disabled,bare-except,ungrouped-imports
try:
    import tensorflow as tf
    tf.autograph.set_verbosity(0)
    from bigml.images.featurizers import ImageFeaturizer as Featurizer
except:
    pass


class Dataset:
    """Local representation of a BigML Dataset. It can store a sample of
    data whose fields are a subset of the ones defined in the fields
    attribute.
    """

    def __init__(self, dataset, api=None, cache_get=None):
        if use_cache(cache_get):
            #pylint: disable=locally-disabled,access-member-before-definition
            self.__dict__ = load(get_dataset_id(dataset), cache_get)
            if self.origin_dataset is not None:
                self.origin_dataset = Dataset(self.origin_dataset,
                    api=api, cache_get=cache_get)
                self.featurizer = Featurizer(self.in_fields,
                    self.input_fields, preferred_only=False)
            return

        self.resource_id = None
        self.name = None
        self.description = None
        self.rows = None
        self.origin_dataset = None
        self.parent_id = None
        self.in_fields = None
        self.out_fields = None
        self.description = None
        self.locale = None
        self.input_fields = None
        self.missing_tokens = None
        self.fields_obj = None
        self.api = get_api_connection(api)
        self.cache_get = cache_get
        self.featurizer = None
        self.transformations = None

        # retrieving dataset information from
        self.resource_id, dataset = get_resource_dict( \
            dataset, "dataset", api=self.api, no_check_fields=False)

        if 'object' in dataset and isinstance(dataset['object'], dict):
            dataset = dataset['object']
            self.name = dataset.get('name')
            self.description = dataset.get('description')
        if 'fields' in dataset and isinstance(dataset['fields'], dict):
            status = get_status(dataset)
            if 'code' in status and status['code'] == FINISHED:
                out_fields_obj = Fields(dataset)
                self.out_fields = out_fields_obj.fields
                self.out_header_names, _ = sorted_headers(out_fields_obj)
                self.out_fields = out_fields_obj.fields
                self.description = dataset["description"]
                self.locale = dataset.get('locale', DEFAULT_LOCALE)
                self.missing_tokens = dataset.get('missing_tokens')
                self.input_fields = dataset.get('input_fields')
                self.rows = dataset.get("rows", 0)
                # we extract the generators and names from the "output_fields"
                if dataset.get("new_fields"):
                    new_fields = get_new_fields(dataset.get(
                        "output_fields", []))
                else:
                    new_fields = None
                origin_dataset = dataset.get("origin_dataset")
                if origin_dataset:
                    self.parent_id = origin_dataset
                    self.add_transformations(origin_dataset, new_fields)
                elif dataset.get("source"):
                    self.parent_id = dataset.get("source")
                    self.in_fields = out_fields_obj.fields
                    self.featurizer = Featurizer(self.in_fields,
                        self.input_fields,
                        self.in_fields,
                        preferred_only=False)
                self.fields_obj = Fields(self.in_fields)
                self.in_header_names, self.in_header_ids = sorted_headers(
                    Fields(self.in_fields))

    def add_transformations(self, origin_dataset, new_fields):
        """Adds a new transformation where the new fields provided are
        defined
        """
        _, origin_dataset = get_resource_dict(
            origin_dataset, "dataset", api=self.api)
        self.origin_dataset = Dataset(origin_dataset, api=self.api,
                                      cache_get=self.cache_get)
        self.in_fields = self.origin_dataset.out_fields
        if new_fields:
            self.transformations = new_fields

    def get_sample(self, rows_number=32):
        """Gets a sample of data representing the dataset """
        sample = self.api.create_sample(self.resource_id)
        if self.api.ok(sample):
            sample = self.api.get_sample(
                sample["resource"], "rows=%s" % rows_number)
            return sample.get("object", {}).get("sample", {}).get("rows")
        return []

    def get_inputs_sample(self, rows_number=32):
        """Gets a sample of data representing the origin dataset """
        if self.origin_dataset is None:
            return []
        return self.origin_dataset.get_sample(rows_number=rows_number)

    def _input_array(self, input_data):
        """Transform the dict-like input data into a row """

        # new_input_data = self.filter_input_data(input_data)
        new_input_data = {}
        for key, value in input_data.items():
            if key not in self.in_fields:
                key = self.fields_obj.fields_by_name.get(key, key)
            new_input_data.update({key: value})
        if self.featurizer is not None:
            new_input_data = self.featurizer.extend_input(new_input_data)
        cast(new_input_data, self.in_fields)
        row = []
        for f_id in self.in_header_ids:
            row.append(None if not f_id in new_input_data else
                new_input_data[f_id])
        return row

    def _transform(self, input_arrays):
        """Given a list of inputs that match the origin dataset structure,
        apply the Flatline transformations used in the dataset

        """
        new_input_arrays = []
        out_headers = []
        fields = {"fields": self.in_fields}
        out_arrays = []
        for transformation in self.transformations:
            expr = transformation.get("field")
            names = transformation.get("names", [])
            out_headers.extend(names)
            # evaluating first to raise an alert if the expression is failing
            check = Flatline.check_lisp(expr, fields)
            if "error" in check:
                raise ValueError(check["error"])
            if expr == '(all)':
                new_input_arrays = input_arrays.copy()
                continue
            new_input = Flatline.apply_lisp(expr, input_arrays, self)
            for index, _ in enumerate(new_input):
                try:
                    new_input_arrays[index]
                except IndexError:
                    new_input_arrays.append([])
                new_input_arrays[index].extend(new_input[index])
        for index, input_array in enumerate(new_input_arrays):
            try:
                out_arrays[index]
            except IndexError:
                out_arrays.append([])
            out_arrays[index].extend(input_array)
        return [out_headers, out_arrays]


    def transform(self, input_data_list):
        """Applies the transformations to the given input data and returns
        the result. Usually, the input_data_list will contain a single
        dictionary, but it can contain a list of them if needed for window
        functions.
        """
        if self.transformations is None and self.featurizer is None:
            return input_data_list
        rows = [self._input_array(input_data) for input_data in
            input_data_list]
        if self.transformations:
            out_headers, out_arrays = self._transform(rows)
            rows = [dict(zip(out_headers, row)) for row
                in out_arrays]
            for index, result in enumerate(rows):
                rows[index] = {key: value for key, value in result.items()
                                  if value is not None}
        else:
            rows = [dict(zip(self.out_header_names, row)) for row in rows]
        return rows

    def dump(self, output=None, cache_set=None):
        """Uses msgpack to serialize the resource object
        If cache_set is filled with a cache set method, the method is called

        """
        self_vars = vars(self).copy()
        del self_vars["api"]
        del self_vars["cache_get"]
        self_vars["origin_dataset"] = self_vars["origin_dataset"].resource_id
        del self_vars["featurizer"]
        del self_vars["fields_obj"]
        dump(self_vars, output=output, cache_set=cache_set)

    def dumps(self):
        """Uses msgpack to serialize the resource object to a string

        """
        self_vars = vars(self).copy()
        del self_vars["api"]
        del self_vars["cache_get"]
        self_vars["origin_dataset"] = self_vars["origin_dataset"].resource_id
        del self_vars["featurizer"]
        del self_vars["fields_obj"]
        return dumps(self_vars)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating public dataset

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create

class TestPublicDataset:
    """Testing published datasets """

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating and reading a public dataset:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I make the dataset public
            And I wait until the dataset is ready less than <dataset_wait> secs
            When I get the dataset status using the dataset's public url
            Then the dataset's status is FINISHED
        """
        show_doc(self.test_scenario1)
        headers = ["data", "source_wait", "dataset_wait"]
        examples = [
            ['data/iris.csv', '10', '10']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file_from_stdin(
                self, example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            dataset_create.make_the_dataset_public(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            dataset_create.build_local_dataset_from_public_url(self)
            dataset_create.dataset_status_finished(self)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,no-member
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from bigml.fields import Fields
from .world import world, eq_, ok_


def i_get_the_missing_values(step):
    """Step: I ask for the missing values counts in the fields"""
    resource = world.dataset
    fields = Fields(resource['fields'])
    step.bigml["result"] = fields.missing_counts()


def i_get_the_errors_values(step):
    """Step: I ask for the error counts in the fields """
    resource = world.dataset
    step.bigml["result"] = world.api.error_counts(resource)


def i_get_the_properties_values(step, properties_dict):
    """Step: the (missing values counts|error counts) dict
    is <properties_dict>
    """
    ok_(properties_dict is not None)
    eq_(step.bigml["result"], json.loads(properties_dict))



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_


def i_check_tst_name(step, name):
    """Step: the statistical test name is <name>"""
    statistical_test_name = world.statistical_test['name']
    eq_(name, statistical_test_name)


def i_create_a_tst_from_dataset(step):
    """Step: I create an statistical test from a dataset"""
    dataset = world.dataset.get('resource')
    resource = world.api.create_statistical_test(dataset, \
        {'name': 'new statistical test'})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.statistical_test = resource['object']
    world.statistical_tests.append(resource['resource'])


def i_update_tst_name(step, name):
    """Step: I update the statistical test name to <name>"""
    resource = world.api.update_statistical_test( \
        world.statistical_test['resource'], {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.statistical_test = resource['object']


def wait_until_tst_status_code_is(step, code1, code2, secs):
    """Step: I wait until the statistical test status code is either
    code1 or code2 less than <secs>"""
    world.statistical_test = wait_until_status_code_is(
        code1, code2, secs, world.statistical_test)


def the_tst_is_finished_in_less_than(step, secs):
    """Step: I wait until the statistical test is ready less than <secs>"""
    wait_until_tst_status_code_is(step, FINISHED, FAULTY, secs)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import json
import os
import zipfile

from bigml.pipeline.pipeline import BMLPipeline, Pipeline
from bigml.api import BigML

from .world import res_filename, eq_, ok_


def i_expand_file_with_models_list(step, pipeline_file, models_list):
    """Extracting models from zip"""
    inner_files = []
    models_list = json.loads(models_list)
    for resource_id in models_list:
        inner_files.append(resource_id.replace("/", "_"))

    pipeline_file = res_filename(pipeline_file)
    with zipfile.ZipFile(pipeline_file, 'r') as zip_ref:
        filenames = [os.path.basename(filename) for
                     filename in zip_ref.namelist()]
        ok_(all(filename in filenames for filename in inner_files))
        zip_ref.extractall(os.path.dirname(pipeline_file))


def i_create_a_local_pipeline_from_models_list(
    step, models_list, name, storage=None):
    """Step: I create a local pipeline for <models_list> named <name> """
    if not isinstance(models_list, list):
        models_list = json.loads(models_list)
    kwargs = {}
    if storage is not None:
        kwargs = {'api': BigML(storage=res_filename(storage))}
    step.bigml["local_pipeline"] = BMLPipeline(name,
                                       models_list,
                                       **kwargs)
    return step.bigml["local_pipeline"]


def the_pipeline_transformed_data_is(step, input_data, output_data):
    """Checking pipeline's transform"""
    if input_data is None:
        input_data = "{}"
    if output_data is None:
        output_data = "{}"
    input_data = json.loads(input_data)
    output_data = json.loads(output_data)
    transformed_data = step.bigml["local_pipeline"].transform([input_data])
    for key, value in transformed_data[0].items():
        eq_(output_data.get(key), value)


def the_pipeline_result_key_is(step, input_data, key, value, precision=None):
    """Checking pipeline transformed property"""
    if input_data is None:
        input_data = "{}"
    input_data = json.loads(input_data)
    transformed_data = step.bigml["local_pipeline"].transform([input_data])
    pipe_value = transformed_data[0].get(key)
    if precision is not None and not isinstance(value, str):
        pipe_value = round(pipe_value, precision)
        value = round(value, precision)
    eq_(str(value), str(pipe_value))


def i_create_composed_pipeline(step, pipelines_list, name):
    """Creating local Pipeline"""
    step.bigml["local_pipeline"] = Pipeline(name, pipelines_list)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member,broad-except
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

from bigml.api import HTTP_CREATED
from bigml.api import FINISHED, FAULTY
from bigml.ensemble import Ensemble
from bigml.ensemblepredictor import EnsemblePredictor
from bigml.model import Model
from bigml.supervised import SupervisedModel
from bigml.local_model import LocalModel

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_

NO_MISSING_SPLITS = {'missing_splits': False}
ENSEMBLE_SAMPLE = {'seed': 'BigML',
                   'ensemble_sample': {"rate": 0.7, "seed": 'BigML'}}


def i_create_an_ensemble(step, number_of_models=2, shared=None):
    """Step: I create an ensemble of <number_of_models> models"""
    if shared is None or world.shared.get("ensemble", {}).get(shared) is None:
        dataset = world.dataset.get('resource')
        try:
            number_of_models = int(number_of_models)
            # tlp is no longer used
            args = {'number_of_models': number_of_models}
        except Exception:
            args = {}
        args.update(NO_MISSING_SPLITS)
        args.update(ENSEMBLE_SAMPLE)
        resource = world.api.create_ensemble(dataset, args=args)
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.ensemble = resource['object']
        world.ensemble_id = resource['resource']
        world.ensembles.append(resource['resource'])


def wait_until_ensemble_status_code_is(step, code1, code2, secs):
    """Step: I wait until the ensemble status code is either <code1> or
    <code2> less than <secs>
    """
    world.ensemble = wait_until_status_code_is(
        code1, code2, secs, world.ensemble)


def the_ensemble_is_finished_in_less_than(step, secs, shared=None):
    """Step: I wait until the ensemble is ready less than <secs>"""
    if shared is None or world.shared.get("ensemble", {}).get(shared) is None:
        wait_until_ensemble_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if "ensemble" not in world.shared:
                world.shared["ensemble"] = {}
            world.shared["ensemble"][shared] = world.ensemble
    else:
        world.ensemble = world.shared["ensemble"][shared]
        world.ensemble_id = world.ensemble["resource"]
        print("Reusing %s" % world.ensemble["resource"])


def create_local_ensemble(step, path=None):
    """Step: I create a local Ensemble"""
    if path is None:
        step.bigml["local_ensemble"] = Ensemble(world.ensemble_id, world.api)
        step.bigml["local_model"] = Model(
            step.bigml["local_ensemble"].model_ids[0], world.api)
    else:
        step.bigml["local_ensemble"] = Ensemble(res_filename(path))
        step.bigml["local_model"] = step.bigml[
            "local_ensemble"].multi_model.models[0]


def create_local_supervised_ensemble(step):
    """Step: I create a local Ensemble"""
    step.bigml["local_ensemble"] = SupervisedModel(world.ensemble_id, world.api)
    step.bigml["local_model"] = Model(step.bigml[
        "local_ensemble"].model_ids[0], world.api)


def create_local_bigml_ensemble(step):
    """Step: I create a local Ensemble"""
    step.bigml["local_ensemble"] = LocalModel(world.ensemble_id, world.api)
    step.bigml["local_model"] = Model(step.bigml[
        "local_ensemble"].model_ids[0], world.api)

def create_local_ensemble_predictor(step, directory):
    """Step: I create a local EnsemblePredictor from <directory>"""
    directory_path = res_filename(directory)
    with open(os.path.join(directory_path, "ensemble.json")) as file_handler:
        ensemble = json.load(file_handler)
    step.bigml["local_ensemble"] = EnsemblePredictor(ensemble, directory)


def load_full_ensemble(step, directory):
    """Step: Given I load the full ensemble information from <directory>"""
    model_list = []
    directory_path = res_filename(directory)
    with open(os.path.join(directory_path, "ensemble.json")) as file_handler:
        ensemble = json.load(file_handler)
        model_list.append(ensemble)
    for model_id in ensemble["object"]["models"]:
        with open(os.path.join(directory_path, model_id.replace("/", "_"))) \
                as file_handler:
            model = json.load(file_handler)
            model_list.append(model)
    return model_list


def create_local_ensemble_with_list(step, number_of_models):
    """Step: I create a local Ensemble with the last <number_of_models>
    models
    """
    step.bigml["local_ensemble"] = Ensemble(world.models[-int(number_of_models):],
                                    world.api)


def create_local_ensemble_from_list(step, model_list):
    """Step: I create a local ensemble from the ensemble <model_list>
    models list
    """
    step.bigml["local_ensemble"] = Ensemble(model_list)


def create_local_ensemble_with_list_of_local_models(step, number_of_models):
    """Step: I create a local Ensemble with the last <number_of_models>
    local models"""
    local_models = [Model(model) for model in
                    world.models[-int(number_of_models):]]
    step.bigml["local_ensemble"] = Ensemble(local_models, world.api)


def field_importance_print(step, field_importance):
    """Step: the field importance text is <field_importance>"""
    field_importance_data = step.bigml["local_ensemble"].field_importance_data()[0]
    eq_(field_importance_data, json.loads(field_importance))


def i_create_an_ensemble_with_params(step, params):
    """Step: I create an ensemble with <params>"""
    dataset = world.dataset.get('resource')
    try:
        args = json.loads(params)
    except Exception:
        args = {}
    args.update(ENSEMBLE_SAMPLE)
    resource = world.api.create_ensemble(dataset, args=args)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.ensemble = resource['object']
    world.ensemble_id = resource['resource']
    world.ensembles.append(resource['resource'])


def i_export_ensemble(step, filename):
    """Step: I export the ensemble"""
    world.api.export(world.ensemble.get('resource'),
                     filename=res_filename(filename))


def i_create_local_ensemble_from_file(step, export_file):
    """Step: I create a local ensemble from file <export_file>"""
    step.bigml["local_ensemble"] = Ensemble(res_filename(export_file))


def check_ensemble_id_local_id(step):
    """Step: the ensemble ID and the local ensemble ID match"""
    eq_(step.bigml["local_ensemble"].resource_id, world.ensemble["resource"])


def clone_ensemble(step, ensemble):
    """Step: I clone ensemble"""
    resource = world.api.clone_ensemble(ensemble,
                                        {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.ensemble = resource['object']
    # save reference
    world.ensembles.append(resource['resource'])


def the_cloned_ensemble_is(step, ensemble):
    """Checking the ensemble is a clone"""
    eq_(world.ensemble["origin"], ensemble)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2012, 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import csv

from bigml.api import HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_, ok_


def i_upload_a_file(step, filename, shared=None):
    """Step: I create a data source uploading a <filename> file"""

    if shared is None or world.shared.get("source", {}).get(shared) is None:
        resource = world.api.create_source(res_filename(filename), \
            {'project': world.project_id})
        # update status
        world.status = resource['code']
        world.location = resource['location']
        world.source = resource['object']
        # save reference
        world.sources.append(resource['resource'])


def i_upload_a_file_with_project_conn(step, filename):
    """Step: I create a data source uploading a <filename> file using
    a project
    """
    resource = world.api.create_source(res_filename(filename))
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])


def i_upload_a_file_from_stdin(step, filename):
    """Step: I create a data source from stdin uploading a <filename> file """
    file_name = res_filename(filename)
    with open(file_name, 'rb') as file_handler:
        resource = world.api.create_source(file_handler, \
            {'project': world.project_id})
        # update status
        world.status = resource['code']
        world.location = resource['location']
        world.source = resource['object']
        # save reference
        world.sources.append(resource['resource'])


def i_upload_a_file_with_args(step, filename, args):
    """Step: I create a data source uploading a <filename> file with args
    <args>
    """
    args = json.loads(args)
    args.update({'project': world.project_id})
    resource = world.api.create_source(res_filename(filename), args)
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])


def i_create_using_url(step, url):
    """Step: I create a data source using the url <url> """
    resource = world.api.create_source(url, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])


def i_create_using_connector(step, connector):
    """Step: I create a data source using the connection <connector>"""
    resource = world.api.create_source(connector, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])


def i_create_composite(step, sources):
    """Step: I create from list of sources <sources>"""
    resource = world.api.create_source(sources, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.composites.append(resource['resource'])


def the_composite_contains(step, sources):
    """Checking source in composite"""
    eq_(world.source["sources"], sources)


def clone_source(step, source):
    """Step: I clone source"""
    resource = world.api.clone_source(source, {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])


def the_cloned_source_origin_is(step, source):
    """Checking cloned source"""
    eq_(world.source["origin"], source)


def i_create_annotated_source(step, directory, args=None):
    """Creating annotated source"""
    if args is None:
        args = {}
    args.update({'project': world.project_id})
    resource = world.api.create_annotated_source(res_filename(directory),
                                                 args)
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.composites.append(resource['resource'])


#pylint: disable=locally-disabled,unnecessary-comprehension
def i_create_using_dict_data(step, data):
    """Step: I create a data source from inline data slurped from <data>"""
    # slurp CSV file to local variable
    mode = 'rt'
    with open(res_filename(data), mode) as fid:
        reader = csv.DictReader(fid)
        dict_data = [row for row in reader]
    # create source
    resource = world.api.create_source(dict_data,
                                       {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.source = resource['object']
    # save reference
    world.sources.append(resource['resource'])


def i_upload_a_file_async(step, filename):
    """Step: I create a data source uploading a <filename> file in
    asynchronous mode
    """
    resource = world.api.create_source(res_filename(filename),
                                       {'project': world.project_id},
                                       async_load=True)
    world.resource = resource


def the_source_has_been_created_async(step, secs):
    """Step: I wait until the source has been created less than <secs> secs"""
    world.source = wait_until_status_code_is(
        FINISHED, FAULTY, secs, world.source)


def wait_until_source_status_code_is(step, code1, code2, secs):
    """Step: I wait until the source status code is either
    <code1> or <code2> less than <secs>
    """
    world.source = wait_until_status_code_is(code1, code2, secs, world.source)


def the_source_is_finished(step, secs, shared=None):
    """Step: I wait until the source is ready less than <secs> """
    if shared is None or world.shared.get("source", {}).get(shared) is None:
        wait_until_source_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if world.shared.get("source") is None:
                world.shared["source"] = {}
            world.shared["source"][shared] = world.source
    else:
        world.source = world.shared["source"][shared]
        print("Reusing %s" % world.source["resource"])


def i_update_source_with(step, data="{}"):
    """Step: I update the source with params <data>"""
    resource = world.api.update_source(world.source.get('resource'), json.loads(data))
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)


def source_has_args(step, args="{}"):
    """Step: the source exists and has args <args>"""
    args = json.loads(args)
    for key, value in list(args.items()):
        if key in world.source:
            eq_(world.source[key], value,
                "Expected key %s: %s. Found %s" % (key, value, world.source[key]))
        else:
            ok_(False, "No key %s in source." % key)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from bigml.api import HTTP_CREATED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_, ok_, approx_


def i_create_a_prediction(step, data=None):
    """Creating prediction"""
    if data is None:
        data = "{}"
    model = world.model['resource']
    data = json.loads(data)
    resource = world.api.create_prediction(model, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_prediction_op(step, data=None, operating_point=None):
    """Creating prediction with operating point"""
    if data is None:
        data = "{}"
    ok_(operating_point is not None)
    model = world.model['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        model, data, {"operating_point": operating_point})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_an_ensemble_prediction_op(step, data=None, operating_point=None):
    """Creating prediction from ensemble with operating point"""
    if data is None:
        data = "{}"
    ok_(operating_point is not None)
    ensemble = world.ensemble['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        ensemble, data, {"operating_point": operating_point})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_fusion_prediction_op(step, data=None, operating_point=None):
    """Create prediction from fusion with operating point"""
    if data is None:
        data = "{}"
    ok_(operating_point is not None)
    fusion = world.fusion['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        fusion, data, {"operating_point": operating_point})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_centroid(step, data=None):
    """Create centroid"""
    if data is None:
        data = "{}"
    cluster = world.cluster['resource']
    data = json.loads(data)
    resource = world.api.create_centroid(cluster, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.centroid = resource['object']
    world.centroids.append(resource['resource'])


def i_create_a_proportional_prediction(step, data=None):
    """Create prediction using proportional strategy for missings"""
    if data is None:
        data = "{}"
    model = world.model['resource']
    data = json.loads(data)
    resource = world.api.create_prediction(model, data,
                                           args={'missing_strategy': 1})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def check_prediction(got, expected, precision=4):
    """Checking prediction is as expected"""
    if not isinstance(got, str):
        approx_(got, float(expected), precision=precision)
    else:
        eq_(got, expected)


def the_prediction_is(step, objective, prediction, precision=4):
    """Checking the prediction for objective field"""
    check_prediction(world.prediction['prediction'][objective], prediction,
                     precision=precision)


def the_median_prediction_is(step, objective, prediction, precision=4):
    """Checking the prediction using median"""
    check_prediction(world.prediction['prediction_path'][
        'objective_summary']['median'], prediction, precision=precision)


def the_centroid_is_with_distance(step, centroid, distance):
    """Checking expected centroid and distance"""
    check_prediction(world.centroid['centroid_name'], centroid)
    check_prediction(world.centroid['distance'], distance)


def the_centroid_is(step, centroid):
    """Checking centroid"""
    check_prediction(world.centroid['centroid_name'], centroid)


def the_centroid_is_ok(step):
    """Checking centroid is ready"""
    ok_(world.api.ok(world.centroid))


def the_confidence_is(step, confidence):
    """Checking confidence"""
    local_confidence = world.prediction.get('confidence', \
        world.prediction.get('probability'))
    approx_(float(local_confidence), float(confidence), precision=4)


def i_create_an_ensemble_prediction(step, data=None):
    """Creating prediction from ensemble"""
    if data is None:
        data = "{}"
    ensemble = world.ensemble['resource']
    data = json.loads(data)
    resource = world.api.create_prediction(ensemble, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_an_ensemble_proportional_prediction(step, data=None, params=None):
    """Creating prediction from ensemble using proportional strategy for
    missings
    """
    if data is None:
        data = "{}"
    if params is None:
        params = {}
    ensemble = world.ensemble['resource']
    data = json.loads(data)
    args = {"missing_strategy": 1}
    args.update(params)
    resource = world.api.create_prediction(ensemble, data, args)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def wait_until_prediction_status_code_is(step, code1, code2, secs):
    """Waiting for prediction and storing result"""
    world.prediction = wait_until_status_code_is(
        code1, code2, secs, world.prediction)


def the_prediction_is_finished_in_less_than(step, secs):
    """Checking wait time"""
    wait_until_prediction_status_code_is(step, FINISHED, FAULTY, secs)


def create_local_ensemble_prediction_add_confidence(step, input_data):
    """Creating prediction from local ensemble with confidence"""
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict(
        json.loads(input_data), full=True)


def create_local_ensemble_prediction(step, input_data):
    """Creating prediction from local ensemble"""
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict(json.loads(input_data))


def create_local_ensemble_prediction_probabilities(step, input_data):
    """Creating prediction from local ensemble with probabilities"""
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict( \
        json.loads(input_data), full=True)
    step.bigml["local_probabilities"] = step.bigml[
        "local_ensemble"].predict_probability( \
        json.loads(input_data), compact=True)


def create_local_ensemble_proportional_prediction_with_confidence( \
    step, input_data, params=None):
    """Creating prediction from local ensemble with confidence"""
    if params is None:
        params = {}
    kwargs = {"full": True, "missing_strategy": 1}
    kwargs.update(params)
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict( \
        json.loads(input_data), **kwargs)

def create_local_ensemble_prediction_using_median_with_confidence( \
    step, input_data):
    """Creating prediction from local ensemble using median with confidence"""
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict( \
        json.loads(input_data), full=True)


def i_create_an_anomaly_score(step, data=None):
    """Creating anomaly score"""
    if data is None:
        data = "{}"
    anomaly = world.anomaly['resource']
    data = json.loads(data)
    resource = world.api.create_anomaly_score(anomaly, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.anomaly_score = resource['object']
    world.anomaly_scores.append(resource['resource'])


def i_create_an_association_set(step, data=None):
    """Creating association set"""
    if data is None:
        data = "{}"
    association = world.association['resource']
    data = json.loads(data)
    resource = world.api.create_association_set(association, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.association_set = resource['object']
    world.association_sets.append(resource['resource'])


def the_anomaly_score_is(step, score):
    """Checking the expected anomaly score"""
    check_prediction(world.anomaly_score['score'], score)


def the_logistic_prediction_is(step, prediction):
    """Checking the expected logistic regression prediction"""
    check_prediction(world.prediction['output'], prediction)


def the_fusion_prediction_is(step, prediction):
    """Checking the expected fusion prediction """
    the_logistic_prediction_is(step, prediction)


def i_create_a_logistic_prediction(step, data=None):
    """Checking the expected logistic regression prediction"""
    if data is None:
        data = "{}"
    model = world.logistic_regression['resource']
    data = json.loads(data)
    resource = world.api.create_prediction(model, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_deepnet_prediction(step, data=None, image_fields=None):
    """Creating a prediction from a deepnet"""
    if data is None:
        data = "{}"
    if image_fields is None:
        image_fields = []
    deepnet = world.deepnet['resource']
    data = json.loads(data)
    data_image_fields = []
    for field in image_fields:
        if field in data:
            data[field] = res_filename(data[field])
            data_image_fields.append(field)
    resource = world.api.create_prediction(deepnet, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    for field in data_image_fields:
        world.sources.append(world.prediction["input_data"][field])
    world.predictions.append(resource['resource'])


def i_create_a_deepnet_prediction_with_op(step, data=None,
                                          operating_point=None):
    """Creating a prediction from a deepnet with operating point"""
    if data is None:
        data = "{}"
    deepnet = world.deepnet['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        deepnet, data, {"operating_point": operating_point})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_logistic_prediction_with_op(step, data=None,
                                           operating_point=None):
    """Creating a prediction from a logistic regression with operating point"""
    if data is None:
        data = "{}"
    logistic_regression = world.logistic_regression['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        logistic_regression, data, {"operating_point": operating_point})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


#pylint: disable=locally-disabled,undefined-loop-variable
def the_logistic_probability_is(step, probability):
    """Checking the logistic regression prediction probability"""
    for [prediction, remote_probability] in world.prediction['probabilities']:
        if prediction == world.prediction['output']:
            break
    approx_(float(remote_probability), float(probability), precision=4)


def the_fusion_probability_is(step, probability):
    """Checking the fusion prediction probability"""
    the_logistic_probability_is(step, probability)


def i_create_a_prediction_op_kind(step, data=None, operating_kind=None):
    """Creating a prediction with operating kind"""
    if data is None:
        data = "{}"
    ok_(operating_kind is not None)
    model = world.model['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        model, data, {"operating_kind": operating_kind})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_an_ensemble_prediction_op_kind(
    step, data=None, operating_kind=None):
    """Creating a prediction from an ensemble with operating kind"""
    if data is None:
        data = "{}"
    ok_(operating_kind is not None)
    ensemble = world.ensemble['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        ensemble, data, {"operating_kind": operating_kind})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_deepnet_prediction_op_kind(step, data=None,
                                          operating_kind=None):
    """Creating a prediction from a deepnet with operating kind"""
    if data is None:
        data = "{}"
    deepnet = world.deepnet['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        deepnet, data, {"operating_kind": operating_kind})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_logistic_prediction_with_op_kind(step, data=None,
                                                operating_kind=None):
    """Creating a prediction from a logistic regression with operating kind"""
    if data is None:
        data = "{}"
    logistic_regression = world.logistic_regression['resource']
    data = json.loads(data)
    resource = world.api.create_prediction( \
        logistic_regression, data, {"operating_kind": operating_kind})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_fusion_prediction(step, data=None):
    """Creating a prediction from a fusion"""
    if data is None:
        data = "{}"
    fusion = world.fusion['resource']
    data = json.loads(data)
    resource = world.api.create_prediction(fusion, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])


def i_create_a_linear_prediction(step, data=None):
    """Creating a prediction from a linear regression"""
    if data is None:
        data = "{}"
    linear_regression = world.linear_regression['resource']
    data = json.loads(data)
    resource = world.api.create_prediction(linear_regression, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.prediction = resource['object']
    world.predictions.append(resource['resource'])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY

from .world import world, eq_
from .read_resource_steps import wait_until_status_code_is


def i_create_configuration(step, configurations):
    """Step: I create a configuration"""
    resource = world.api.create_configuration(
        configurations, {"name": "configuration"})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.configuration = resource['object']
    world.configurations.append(resource['resource'])


def i_update_configuration(step, changes):
    """Step: I update a configuration"""
    resource = world.api.update_configuration(
        world.configuration["resource"], changes)
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.configuration = resource['object']


def wait_until_configuration_status_code_is(step, code1, code2, secs):
    """Step: I wait until the configuration status code is either <code1> or
    <code2> less than <secs>
    """
    world.configuration = wait_until_status_code_is(
        code1, code2, secs, world.configuration)


def the_configuration_is_finished_in_less_than(step, secs):
    """Step: I wait until the configuration is ready less than <secs>"""
    wait_until_configuration_status_code_is(step, FINISHED, FAULTY, secs)


def i_check_configuration_name(step, name):
    """Step: the configuration name is <name>"""
    eq_(world.configuration["name"], name["name"])


def i_check_configuration_conf(step, confs):
    """Step: the configuration contents are <confs>"""
    eq_(world.configuration["configurations"], confs)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_


def the_library_code_and_attributes(step, source_code, param, param_value):
    """Step: the library code is <source_code> and the value of <param>
    is <param_value>
    """
    res_param_value = world.library[param]
    eq_(res_param_value, param_value,
        ("The library %s is %s and the expected %s is %s" %
         (param, param_value, param, param_value)))


def i_create_a_library(step, source_code):
    """Step: I create a whizzml library from a excerpt of code <source_code>"""
    resource = world.api.create_library(source_code,
                                        {"project": world.project_id})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.library = resource['object']
    world.libraries.append(resource['resource'])


def i_update_a_library(step, param, param_value):
    """Step: I update the library with <param>, <param_value>"""
    resource = world.api.update_library(world.library['resource'],
                                        {param: param_value})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.library = resource['object']


def wait_until_library_status_code_is(step, code1, code2, secs):
    """Step: I wait until the library status code is either <code1> or
    <code2> less than <secs>
    """
    world.library = wait_until_status_code_is(
        code1, code2, secs, world.library)


def the_library_is_finished(step, secs):
    """Step: I wait until the library is ready less than <secs>"""
    wait_until_library_status_code_is(step, FINISHED, FAULTY, secs)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2012, 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import json

import bigml.generators.model as g

from bigml.tests.world import res_filename
from bigml.predict_utils.common import extract_distribution
from bigml.util import utf8

from .world import world, eq_


#pylint: disable=locally-disabled,invalid-name
def i_translate_the_tree_into_IF_THEN_rules(step):
    """Step: I translate the tree into IF-THEN rules"""
    output = io.StringIO()
    g.rules(step.bigml["local_model"], out=output)
    world.output = output.getvalue()

def i_check_the_data_distribution(step, filename):
    """Step: I check data distribution with <filename> file"""
    distribution = g.get_data_distribution(step.bigml["local_model"])

    distribution_str = ''
    for bin_value, bin_instances in distribution:
        distribution_str += "[%s,%s]\n" % (bin_value, bin_instances)
    world.output = utf8(distribution_str)
    i_check_if_the_output_is_like_expected_file(step, filename)


def i_check_the_predictions_distribution(step, filename):
    """Step: I check the predictions distribution with <filename> file"""
    predictions = g.get_prediction_distribution(step.bigml["local_model"])

    distribution_str = ''
    for group, instances in predictions:
        distribution_str += "[%s,%s]\n" % (group, instances)

    world.output = utf8(distribution_str)

    i_check_if_the_output_is_like_expected_file(step, filename)


def i_check_the_model_summary_with(step, filename):
    """Step: I check the model summary with <filename> file"""
    output = io.StringIO()
    g.summarize(step.bigml["local_model"], out=output)
    world.output = output.getvalue()
    i_check_if_the_output_is_like_expected_file(step, filename)


def i_check_if_the_output_is_like_expected_file(step, expected_file):
    """Step: I check the output is like <expected_file> expected file"""
    with open(res_filename(expected_file), "r") as handler:
        expected_content = handler.read()
    eq_(world.output.strip(), expected_content.strip())


def i_check_print_distribution(step, filename):
    """Step: I check the distribution print with <filename> file"""
    output = io.StringIO()
    _, distribution = extract_distribution(
        step.bigml["local_model"].root_distribution)
    g.print_distribution(distribution, output)
    world.output = output.getvalue()
    if world.debug:
        backup = "%s.bck" % filename
        with open(backup, "w") as bck_file:
            bck_file.write(world.output)
    i_check_if_the_output_is_like_expected_file(step, filename)


def i_list_fields(step, filename):
    """Step: I check the list fields print with <filename> file"""
    output = io.StringIO()
    g.list_fields(step.bigml["local_model"], output)
    world.output = output.getvalue()
    if world.debug:
        backup = "%s.bck" % filename
        with open(backup, "w") as bck_file:
            bck_file.write(world.output)
    i_check_if_the_output_is_like_expected_file(step, filename)


def i_create_tree_csv(step, filename):
    """Step: I check the tree csv print with <filename> file"""
    rows = g.tree_csv(step.bigml["local_model"])
    world.output = json.dumps(rows)
    if world.debug:
        backup = "%s.bck" % filename
        with open(backup, "w") as bck_file:
            bck_file.write(world.output)
    i_check_if_the_output_is_like_expected_file(step, filename)

def update_content(filename, content):
    """Step: I check the tree csv print with <filename> file"""
    with open(res_filename(filename), "w") as file_handler:
        file_handler.write(content)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json

from bigml.dataset import Dataset

from .world import res_filename, eq_


def i_create_a_local_dataset_from_file(step, dataset_file):
    """Step: I create a local dataset from a <dataset_file> file"""
    step.bigml["local_dataset"] = Dataset(res_filename(dataset_file))


def the_transformed_data_is(step, input_data, output_data):
    """Checking expected transformed data"""
    if input_data is None:
        input_data = "{}"
    if output_data is None:
        output_data = "{}"
    input_data = json.loads(input_data)
    output_data = json.loads(output_data)
    transformed_data = step.bigml["local_dataset"].transform([input_data])
    for key, value in transformed_data[0].items():
        eq_(output_data.get(key), value)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_


def i_check_sample_name(step, name):
    """Step: the sample name is <name>"""
    sample_name = world.sample['name']
    eq_(name, sample_name)


def i_create_a_sample_from_dataset(step):
    """Step: I create a sample from a dataset"""
    dataset = world.dataset.get('resource')
    resource = world.api.create_sample(dataset, {'name': 'new sample'})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.sample = resource['object']
    world.samples.append(resource['resource'])


def i_update_sample_name(step, name):
    """Step: I update the sample name to <name>"""
    resource = world.api.update_sample(world.sample['resource'],
                                       {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.sample = resource['object']


def the_sample_is_finished_in_less_than(step, secs):
    """Step: I wait until the sample is ready less than <secs>"""
    world.sample = wait_until_status_code_is(
        FINISHED, FAULTY, secs, world.sample)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating local ensemble predictions

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_model_steps as model_create
from . import create_ensemble_steps as ensemble_create
from . import create_prediction_steps as prediction_create
from . import compare_predictions_steps as compare_pred


class TestEnsemblePrediction:
    """Testing local ensemble prediction"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating a local prediction from an Ensemble:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create an ensemble of <number_of_models> models
            And I wait until the ensemble is ready less than <model_wait> secs
            And I create a local Ensemble
            When I create a local ensemble prediction with probabilities for "<input_data>"
            Then the local prediction is "<prediction>"
            And the local prediction's confidence is "<confidence>"
            And the local probabilities are "<probabilities>"
        """
        show_doc(self.test_scenario1)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "number_of_models", "input_data",
                   "prediction", "confidence", "probabilities"]
        examples = [
            ['data/iris.csv', '10', '10', '50', '5',
             '{"petal width": 0.5}', 'Iris-versicolor', '0.415',
             '["0.3403", "0.4150", "0.2447"]' ]]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            ensemble_shared = "%s_%s" % (example["data"],
                example["number_of_models"])
            ensemble_create.i_create_an_ensemble(
                self, example["number_of_models"], shared=ensemble_shared)
            ensemble_create.the_ensemble_is_finished_in_less_than(
                self, example["model_wait"], shared=ensemble_shared)
            ensemble_create.create_local_ensemble(self)
            prediction_create.create_local_ensemble_prediction_probabilities(
                self, example["input_data"])
            compare_pred.the_local_prediction_is(self, example["prediction"])
            compare_pred.the_local_prediction_confidence_is(
                self, example["confidence"])
            compare_pred.the_local_probabilities_are(
                self, example["probabilities"])

    def test_scenario2(self):
        """
        Scenario: Successfully obtaining field importance from an Ensemble:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a model with "<model_conf1>"
            And I wait until the model is ready less than <model_wait> secs
            And I create a model with "<model_conf2>"
            And I wait until the model is ready less than <model_wait> secs
            And I create a model with "<model_conf3>"
            And I wait until the model is ready less than <model_wait> secs
            When I create a local Ensemble with the last <number_of_models> models
            Then the field importance text is <field_importance>
        """
        show_doc(self.test_scenario2)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "model_conf1", "model_conf2", "model_conf3",
                   "number_of_models", "field_importance"]
        examples = [
            ['data/iris.csv', '10', '10', '20',
             '{"input_fields": ["000000", "000001","000003", "000004"]}',
             '{"input_fields": ["000000", "000001","000002", "000004"]}',
             '{"input_fields": ["000000", "000001","000002", "000003",'
             ' "000004"]}', '3',
             '[["000002", 0.5269933333333333], ["000003", 0.38936],'
             ' ["000000", 0.04662333333333333],'
             '["000001", 0.037026666666666666]]']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_model_with(self, example["model_conf1"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            model_create.i_create_a_model_with(self, example["model_conf2"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            model_create.i_create_a_model_with(self, example["model_conf3"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            ensemble_create.create_local_ensemble_with_list(
                self, example["number_of_models"])
            ensemble_create.field_importance_print(
                self, example["field_importance"])

    def test_scenario3(self):
        """
        Scenario: Successfully creating a local prediction from an Ensemble adding confidence:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create an ensemble of <number_of_models> models
            And I wait until the ensemble is ready less than <model_wait> secs
            And I create a local Ensemble
            When I create a local ensemble prediction for "<input_data>" in JSON adding confidence
            Then the local prediction is "<prediction>"
            And the local prediction's confidence is "<confidence>"
        """
        show_doc(self.test_scenario3)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "number_of_models", "input_data", "prediction",
                   "confidence"]
        examples = [
            ['data/iris.csv', '10', '10', '50', '5',
             '{"petal width": 0.5}', 'Iris-versicolor', '0.415']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            ensemble_shared = "%s_%s" % (example["data"],
                example["number_of_models"])
            ensemble_create.i_create_an_ensemble(
                self, example["number_of_models"], shared=ensemble_shared)
            ensemble_create.the_ensemble_is_finished_in_less_than(
                self, example["model_wait"], shared=ensemble_shared)
            ensemble_create.create_local_ensemble(self)
            prediction_create.create_local_ensemble_prediction_add_confidence(
                self, example["input_data"])
            compare_pred.the_local_prediction_is(self, example["prediction"])
            compare_pred.the_local_prediction_confidence_is(
                self, example["confidence"])

    def test_scenario4(self):
        """
        Scenario: Successfully obtaining field importance from an Ensemble created from local models:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a model with "<model_conf1>"
            And I wait until the model is ready less than <model_wait> secs
            And I create a model with "<model_conf2>"
            And I wait until the model is ready less than <model_wait> secs
            And I create a model with "<model_conf3>"
            And I wait until the model is ready less than <model_wait> secs
            When I create a local Ensemble with the last <number_of_models> local models
            Then the field importance text is <field_importance>
        """
        show_doc(self.test_scenario4)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "model_conf1", "model_conf2", "model_conf3",
                   "number_of_models", "field_importance"]
        examples = [
            ['data/iris.csv', '10', '10', '30',
             '{"input_fields": ["000000", "000001","000003", "000004"]}',
             '{"input_fields": ["000000", "000001","000002", "000004"]}',
             '{"input_fields": ["000000", "000001","000002", "000003",'
             ' "000004"]}', '3',
             '[["000002", 0.5269933333333333], ["000003", 0.38936],'
             ' ["000000", 0.04662333333333333], '
             '["000001", 0.037026666666666666]]']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_model_with(self, example["model_conf1"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            model_create.i_create_a_model_with(self, example["model_conf2"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            model_create.i_create_a_model_with(self, example["model_conf3"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            ensemble_create.create_local_ensemble_with_list_of_local_models(
                self, example["number_of_models"])
            ensemble_create.field_importance_print(
                self, example["field_importance"])

    def test_scenario5(self):
        """
        Scenario: Successfully creating a local prediction from an Ensemble:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create an ensemble of <number_of_models> models
            And I wait until the ensemble is ready less than <model_wait> secs
            And I create a local Ensemble
            When I create a local ensemble prediction using median with confidence for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        show_doc(self.test_scenario5)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "number_of_models", "input_data", "prediction"]
        examples = [
            ['data/grades.csv', '30', '30', '50', '2', '{}', 69.0934]]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            ensemble_shared = "%s_%s" % (example["data"],
                example["number_of_models"])
            ensemble_create.i_create_an_ensemble(
                self, example["number_of_models"], shared=ensemble_shared)
            ensemble_create.the_ensemble_is_finished_in_less_than(
                self, example["model_wait"], shared=ensemble_shared)
            ensemble_create.create_local_ensemble(self)
            prediction_create.create_local_ensemble_prediction_using_median_with_confidence(
                self, example["input_data"])
            compare_pred.the_local_prediction_is(self, example["prediction"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2020-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from bigml.api import HTTP_ACCEPTED
from bigml.api import FINISHED
from bigml.api import FAULTY

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_, ok_


def i_create_external_connector(step):
    """Step: I create an external connector"""
    resource = world.api.create_external_connector(None, \
        {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.external_connector = resource['object']
    # save reference
    world.external_connectors.append(resource['resource'])


def wait_until_external_connector_status_code_is(step, code1, code2, secs):
    """Step: I wait until the external connector status code is either
    <code1> or <code2> less than <step>
    """
    world.external_connector = wait_until_status_code_is(
        code1, code2, secs, world.external_connector)


def the_external_connector_is_finished(step, secs):
    """Step: I wait until the external_connector is ready less than <secs> """
    wait_until_external_connector_status_code_is(step, FINISHED, FAULTY, secs)


def i_update_external_connector_with(step, data="{}"):
    """Step: I update the external_connector with params <data>"""
    resource = world.api.update_external_connector( \
        world.external_connector.get('resource'), json.loads(data))
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)


def external_connector_has_args(step, args="{}"):
    """Step: the external connector exists and has args <args>"""
    args = json.loads(args)
    for key, value in list(args.items()):
        if key in world.external_connector:
            eq_(world.external_connector[key], value,
                "Expected key %s: %s. Found %s" % (key, value, world.external_connector[key]))
        else:
            ok_(False, "No key %s in external connector." % key)



# -*- coding: utf-8 -*-
#pylint: disable=invalid-name
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Image Featurizers

This module defines the classes that produce the features extracted from
images in BigML. They are used in Modefields to extend the original input
data provided for local predictions.

"""
import os
import math
import numpy as np


from PIL import Image
from sensenet.models.wrappers import create_image_feature_extractor
from bigml.featurizer import Featurizer, expand_date
from bigml.constants import IMAGE

TOP_SIZE = 512
N_BINS = 16
INTENSITY_RANGE = 256
BIN_WIDTH = INTENSITY_RANGE / N_BINS
HOG_BINS = 9
HOG_BIN_WIDTH = np.pi / HOG_BINS
DECOMPS = ["horizontal", "diagonal", "vertical"]

PRETRAINED = "pretrained_cnn"
WAVELET = "wavelet_subbands"

def resize_to(image, top_size=TOP_SIZE):
    """Resizing the image to a maximum width or height """
    width, height = image.size
    if width > top_size or height > top_size:
        if width > height:
            ratio = height / width
            image = image.resize((top_size , int(ratio * top_size)),
                                 Image.BICUBIC)
        else:
            ratio = width / height
            image = image.resize((int(ratio * top_size), top_size),
                                  Image.BICUBIC)
    return image


def grid_coords(image_a, grid_size):
    """ getting the start and end positions for each grid """
    try:
        height, width, _ = image_a.shape
    except ValueError:
        height, width = image_a.shape
    f_grid_size = float(grid_size)
    h_step = height / f_grid_size
    w_step = width / f_grid_size
    coords = []
    for h in range(0, grid_size):
        for w in range(0, grid_size):
            h_start = int(max([0, math.floor(h * h_step)]))
            w_start = int(max([0, math.floor(w * w_step)]))
            h_end = int(min([height, math.ceil((h + 1) * h_step)]))
            w_end = int(min([width, math.ceil((w + 1) * w_step)]))
            coords.append([h_start, w_start, h_end, w_end])
    return coords


def dimensions_extractor(image_file):
    """Returns the features related to the image dimensions:
       file size, width, height, aspect ratio
    """
    file_size = os.stat(image_file).st_size
    image = Image.open(image_file)
    width, height = image.size
    aspect_ratio = width / float(height)
    return [file_size, width, height, aspect_ratio]


def average_pixels_extractor(image_file):
    """ Averaging pixels for the entire image, 3x3 and 4x4 grids
    The image passed as argument should already be resized to 512 max
    """
    image = Image.open(image_file)
    image = resize_to(image)
    image_a = np.array(image)
    avg_pixels =  [np.average(image_a[:, :, n]) for n in range(0, 3)]
    coords = grid_coords(image_a, 3)
    coords.extend(grid_coords(image_a, 4))
    for h_start, w_start, h_end, w_end in coords:
        avg_pixels.extend(
            [np.average(image_a[h_start: h_end, w_start: w_end, n])
             for n in range(0, 3)])
    return avg_pixels


def get_bin(value, bin_width):
    """Returns the bin where a value falls in."""
    return math.floor(value / bin_width)


def get_luminance(image_a):
    """Getting the Y coordinate in YUV in terms of the RGB channel info."""
    r = image_a[:, :, 0]
    g = image_a[:, :, 1]
    b = image_a[:, :, 2]

    image_l = 0.299 * r + 0.587 * g + 0.114 * b
    image_l = image_l.astype('d')
    return image_l

def level_histogram_extractor(image_file):
    """Level histogram feature extractor."""
    image = Image.open(image_file)
    image = resize_to(image)
    image_a = np.array(image)
    height, width, _ = image_a.shape
    pixels_per_channel = width * height
    output = [0] * 3 * N_BINS
    for c in range(0, 3):
        offset = N_BINS * c
        for h in range(0, height):
            for w in range(0, width):
                bin_index = get_bin(image_a[h][w][c], BIN_WIDTH)
                output[bin_index + offset] += 1
    for index, _ in enumerate(output):
        output[index] /= pixels_per_channel

    return output


def HOG_transform(image_a):
    """Histogram of Gradients transformation."""
    image_l = get_luminance(image_a)
    height, width = image_l.shape
    if height > 2 and width > 2:
        trans_image = np.empty(((height - 2), (width - 2), 2))
        trans_image.astype('d')
        for y in range(0, (height - 2)):
            for x in range(0, (width - 2)):
                py = y + 1
                px = x + 1
                x_edge = image_l[py][x] - image_l[py][px + 1]
                y_edge = image_l[y][px] - image_l[py + 1][px]

                trans_image[y][x][0] = math.sqrt(
                    x_edge * x_edge + y_edge * y_edge)

                # Convert to zero - pi radians
                if x_edge == 0:
                    if y_edge > 0:
                        trans_image[y][x][1] = np.pi
                    elif y_edge < 0:
                        trans_image[y][x][1] = 0
                    else:
                        trans_image[y][x][1] = np.nan
                else:
                    trans_image[y][x][1] = math.atan(
                        y_edge / x_edge) + (np.pi / 2)
    else:
        trans_image = np.empty((height, width, 2))
        for y in range(0, height):
            for x in range(0, width):
                trans_image[y][x][0] = 0
                trans_image[y][x][1] = np.nan

    return trans_image


def HOG_aggregate(trans_image, grid_size):
    """Histogram of Gradients aggregation."""
    # Laplace correction to avoid zero norm; kind of arbitrary
    features = np.ones(((grid_size * grid_size), HOG_BINS))

    bounds = grid_coords(trans_image, grid_size)
    for index, bound in enumerate(bounds):
        h_start, w_start, h_end, w_end = bound
        for y in range(h_start, h_end):
            for x in range(w_start, w_end):
                mag = trans_image[y][x][0]
                angle = trans_image[y][x][1]

                if mag > 0:
                    if angle >= np.pi:
                        low = HOG_BINS - 1
                    else:
                        low = get_bin(angle, HOG_BIN_WIDTH)
                    high = (low + 1) % HOG_BINS

                    high_weight = (
                        angle - low * HOG_BIN_WIDTH) / HOG_BIN_WIDTH
                    low_weight = 1 - high_weight

                    # Split vote between adjacent bins
                    features[index][low] += mag * low_weight
                    features[index][high] += mag * high_weight
        norm = np.linalg.norm(features[index])
        features[index] = features[index] / norm
    return features


def HOG_extractor(image_file):
    """Histogram of Gradients Feature extractor"""
    image = Image.open(image_file)
    image = image.convert('RGB')
    image = resize_to(image)
    image_a = np.array(image)
    transform = HOG_transform(image_a)
    features = HOG_aggregate(transform, 1)
    features3x3 = HOG_aggregate(transform, 3)
    features4x4 = HOG_aggregate(transform, 4)
    features_list = list(features.reshape(-1))
    features_list.extend(list(features3x3.reshape(-1)))
    features_list.extend(list(features4x4.reshape(-1)))
    return features_list


def energy_parameters(values, coords):
    """Energy parameters computation."""
    if len(values) < 2 and len(values[0]) < 2:
        return np.array([values[0][0], 0])
    count = 0
    mean = 0
    sum_sq = 0
    h_start, w_start, h_end, w_end = coords

    for y in range(h_start, h_end):
        for x in range(w_start, w_end):
            new_value = values[y][x]
            count += 1
            delta1 = new_value - mean
            mean += delta1 / count
            delta2 = new_value - mean
            sum_sq += delta1 * delta2

    return np.array([mean, sum_sq / (count - 1)])


def haar1Ds(signal):
    """1-dimensional Haard components."""
    output = np.empty((2, max([1, int(len(signal) / 2)])))

    if len(signal) > 1:
        for i in range(0, len(signal) - 1, 2):
            index = int(i / 2)
            output[0][index] = (signal[i] + signal[i + 1]) / 2
            output[1][index] = abs(signal[i] - signal[i + 1])

    else:
        output[0][0] = signal[0]
        output[1][0] = 0

    return output


def haar1D(image, vertical):
    """1-dimensional Haard vertical component."""
    if vertical:
        image = image.transpose()

    output = np.empty((2, len(image), max([1, int(len(image[0]) / 2)])))

    for i, cell in enumerate(image):
        row_decomp = haar1Ds(cell)
        output[0][i] = row_decomp[0]
        output[1][i] = row_decomp[1]

    if vertical:
        output = np.array([output[0].transpose(),
                           output[1].transpose()])

    return output


def haar2D(image):
    """2-dimensional Haard components."""
    h_mean, h_detail = haar1D(image, False)
    average, vertical = haar1D(h_mean, True)
    horizontal, diagonal = haar1D(h_detail, True)

    return np.array([average, horizontal, diagonal, vertical])


def wavelet_subbands_aggregate(trans_image, grid_size):
    """Wavelet subbands aggregation. """
    index = 0
    features = np.empty((((len(trans_image) - 1) * len(DECOMPS) + 1) *
        grid_size * grid_size * 2,))
    features.astype('d')
    bounds = []
    for cell in trans_image:
        bounds.append(grid_coords(cell[0], grid_size))
    for cell_index in range(grid_size * grid_size):
        for i, row in enumerate(trans_image):
            for cell in row:
                params = energy_parameters(
                    cell, bounds[i][cell_index])
                features[index] = params[0]
                features[index + 1] = params[1]

                index += len(params)

    return features


def wavelet_subbands_transform(image_a, levels):
    """Haard Wavelet subbands transformation."""
    image_l = get_luminance(image_a)

    output = []

    for _ in range(0, levels):
        level_output = []
        decomp = haar2D(image_l)
        for j in range(0, len(DECOMPS)):
            level_output.append(decomp[j + 1])
        image_l = decomp[0]
        output.append(level_output)

    output.append([image_l])

    return output


def wavelet_subbands_extractor(image_file, levels):
    """Wavelet subbands feature extractor."""
    image = Image.open(image_file)
    image = image.convert('RGB')
    image = resize_to(image)
    image_a = np.array(image)
    transform = wavelet_subbands_transform(image_a, levels)
    features = wavelet_subbands_aggregate(transform, 1)
    features2x2 = wavelet_subbands_aggregate(transform, 2)
    features_list = list(features.reshape(-1))
    features_list.extend(list(features2x2.reshape(-1)))
    return features_list


IMAGE_EXTRACTORS = {
    "dimensions": dimensions_extractor,
    "average_pixels": average_pixels_extractor,
    "level_histogram": level_histogram_extractor,
    "histogram_of_gradients": HOG_extractor
}

IMAGE_PROVENANCE = list(IMAGE_EXTRACTORS.keys()) + [PRETRAINED, WAVELET]


#pylint: disable=locally-disabled,bare-except
def get_image_extractors(res_object, field_id):
    """Returns the feature extractor function for an image field"""
    extractors = []
    try:
        extracted_features = res_object.fields[field_id].get(
            "image_analysis", {}).get("extracted_features")
        for feature in extracted_features:
            if isinstance(feature, list) and feature[0] == PRETRAINED:
                _, cnn_name = feature[:]
                extractors.append(lambda x, param=cnn_name: list(
                    create_image_feature_extractor(param, None)(x))[0])
            elif isinstance(feature, list) and feature[0] == WAVELET:
                _, levels = feature[:]
                extractors.append(lambda x, param=levels:
                    wavelet_subbands_extractor(x, param))
            else:
                extractors.append(IMAGE_EXTRACTORS[feature])

    except:
        pass
    return extractors


def expand_image(res_object, parent_id, image_file):
    """ Retrieves all the values of the subfields generated from
    a parent image field

    """
    expanded = {}
    keys = res_object.fields[parent_id]["child_ids"]
    values = []
    for generator in res_object.generators[parent_id]:
        values.extend(generator(image_file))
    expanded = dict(zip(keys, values))
    return expanded


class ImageFeaturizer(Featurizer):
    """This class provides methods for image Feature extraction."""

    def __init__(self, fields, input_fields, selected_fields=None,
                 preferred_only=True):
        self.fields = fields
        self.input_fields = input_fields
        self.subfields = {}
        self.generators = {}
        self.preferred_only = preferred_only
        self.selected_fields = self.add_subfields(selected_fields,
            preferred_only=preferred_only)
        super().__init__(fields, input_fields, selected_fields, preferred_only)

    def _add_subfield(self, field_id, field):
        """Adding a subfield and the corresponding generator """
        parent_id = field["parent_ids"][0]
        subfield = {field_id: field["datatype"]}
        if parent_id in list(self.subfields.keys()):
            self.subfields[parent_id].update(subfield)
        else:
            parent_type = self.fields[parent_id]["optype"]
            expand_fn_list = get_image_extractors(self, parent_id) \
                if parent_type == IMAGE else [expand_date]
            self.selected_fields[parent_id] = self.fields[parent_id]
            self.subfields[parent_id] = subfield
            self.generators.update({parent_id: expand_fn_list})

    def add_subfields(self, selected_fields=None, preferred_only=True):
        """Adding the subfields information in the fields structure and the
        generating functions for the subfields values.
        """
        # filling preferred fields with preferred input fields
        fields = selected_fields or self.fields

        if selected_fields is None:
            selected_fields = {}
            selected_fields.update({field_id: field for field_id, field \
                in fields.items() if field_id in self.input_fields \
                and (not preferred_only or self.fields[field_id].get(
                "preferred", True))})
        self.selected_fields = selected_fields

        # computing the generated subfields
        for fid, finfo in list(self.selected_fields.items()):
            if finfo.get('parent_optype', False) == 'datetime' or \
                finfo.get('provenance', False) in IMAGE_PROVENANCE:
                # datetime and image subfields
                self._add_subfield(fid, finfo)

        return self.selected_fields

    def extend_input(self, input_data):
        """Computing the values for the generated subfields and adding them
        to the original input data. Parent fields will be removed.
        """
        extended = {}
        for f_id, value in list(input_data.items()):
            if f_id in self.generators.keys():
                if not self.preferred_only:
                    extended[f_id] = value
                if self.fields[f_id]["optype"] == IMAGE:
                    extended.update(expand_image(self, f_id, input_data[f_id]))
                else:
                    extended.update(
                        self.generators[f_id][0](self, f_id, input_data[f_id]))
            else:
                extended[f_id] = value
        return extended



# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A Featurizer to generate features for composed fields.

This module defines a Featurizer class to hold the information associated
to the subfields derived from datetime fields.
It is used for local predictions.

"""
from bigml_chronos import chronos
from bigml.constants import DATETIME


DATE_FNS = {
    "day-of-month": lambda x: x.day,
    "day-of-week": lambda x: x.weekday() + 1,
    "millisecond": lambda x: x.microsecond / 1000}

IMAGE_PROVENANCE = ["dimensions", "average_pixels", "level_histogram",
    "histogram_of_gradients", "pretrained_cnn", "wavelet_subbands"]

def expand_date(res_object, parent_id, date):
    """ Retrieves all the values of the subfields generated from
    a parent datetime field

    """
    expanded = {}
    timeformats = res_object.fields[parent_id].get('time_formats', {})
    try:
        parsed_date = chronos.parse(date, format_names=timeformats)
    except ValueError:
        return {}
    for fid, ftype in list(res_object.subfields[parent_id].items()):
        date_fn = DATE_FNS.get(ftype)
        if date_fn is not None:
            expanded.update({fid: date_fn(parsed_date)})
        else:
            expanded.update({fid: getattr(parsed_date, ftype)})
    return expanded


class Featurizer:
    """A class to generate the components derived from a composed field """

    def __init__(self, fields, input_fields, selected_fields=None,
                 preferred_only=True):
        self.fields = fields
        self.input_fields = input_fields
        self.subfields = {}
        self.generators = {}
        self.preferred_only = preferred_only
        self.selected_fields = self.add_subfields(
            selected_fields, preferred_only=preferred_only)

    def add_subfields(self, selected_fields=None, preferred_only=True):
        """Adding the subfields information in the fields structure and the
        generating functions for the subfields values.
        """
        # filling preferred fields with preferred input fields
        fields = selected_fields or self.fields

        if selected_fields is None:
            selected_fields = {}
            selected_fields.update({field_id: field for field_id, field \
                in fields.items() if field_id in self.input_fields \
                and (not preferred_only or self.fields[field_id].get(
                "preferred", True))})

        # computing the subfields generated from parsing datetimes
        for fid, finfo in list(selected_fields.items()):

            # datetime subfields
            if finfo.get('parent_optype', False) == DATETIME:
                parent_id = finfo["parent_ids"][0]
                subfield = {fid: finfo["datatype"]}
                if parent_id in list(self.subfields.keys()):
                    self.subfields[parent_id].update(subfield)
                else:
                    selected_fields[parent_id] = self.fields[parent_id]
                    self.subfields[parent_id] = subfield
                    self.generators.update({parent_id: expand_date})
            elif finfo.get('provenance', False) in IMAGE_PROVENANCE:
                raise ValueError("This model uses image-derived fields. "
                                 "Please, use the pip install bigml[images] "
                                 "option to install the libraries required "
                                 "for local predictions in this case.")

        return selected_fields

    def extend_input(self, input_data):
        """Computing the values for the generated subfields and adding them
        to the original input data. Parent fields will be removed if the
        `preferred_only` option is set, as they are not used in models.
        However, the `preferred_only` option set to False will keep them,
        allowing to be used as generators in other transformations.
        """
        extended = {}
        for f_id, value in list(input_data.items()):
            if f_id in self.subfields:
                if not self.preferred_only:
                    extended[f_id] = value
                extended.update(self.generators[f_id](self, f_id, value))
            else:
                extended[f_id] = value
        return extended



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2016-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating a local Topic distribution from Topic Model

"""
import sys

from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_lda_steps as topic_create
from . import compute_lda_prediction_steps as lda_predict


# This model is from the bigmlcom/streaming-lda; the associated test is
# for near-exact equivalence with that library (with special attention
# to random number generation).
DUMMY_MODEL = {
    "input_fields": ["000001"],
    "topic_model": {
        "alpha": 0.08,
        "beta": 0.1,
        "hashed_seed": 0,
        "language": "en",
        "bigrams": True,
        "case_sensitive": False,
        "term_topic_assignments": [[0, 0, 1, 2],
                                   [0, 1, 2, 0],
                                   [1, 2, 0, 0],
                                   [0, 0, 2, 0]],
        "termset": ["cycling", "playing", "shouldn't", "uńąnimous court"],
        "options": {},
        "topics": [{"name": "Topic 1",
                    "id": "000000",
                    "top_terms": ["a", "b"],
                    "probability": 0.1},
                   {"name": "Topic 2",
                    "id": "000001",
                    "top_terms": ["c", "d"],
                    "probability": 0.1},
                   {"name": "Topic 3",
                    "id": "000000",
                    "top_terms": ["e", "f"],
                    "probability": 0.1},
                   {"name": "Topic 4",
                    "id": "000000",
                    "top_terms": ["g", "h"],
                    "probability": 0.1}],
        "fields": {
            "000001": {
                "datatype": "string",
                "name": "TEST TEXT",
                "optype": "text",
                "order": 0,
                "preferred": True,
                "summary": {},
                "term_analysis": {}
            }
        }
    },
    "resource": "topicmodel/aaaaaabbbbbbccccccdddddd"
}


class TestTopicModel:
    """Test Topic Model Predictions"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario 1: Successfully creating a local Topic Distribution
            Given I have a block of text and an LDA model
            And I use the model to predict the topic distribution
            Then the value of the distribution matches the expected distribution
        """
        show_doc(self.test_scenario1)
        headers = ["model", "text", "expected_distribution"]
        examples = [
            # This example is a replication of a test in bigmlcom/streaming-lda
            [
                DUMMY_MODEL,
                {"TEST TEXT": "uńąnimous court 'UŃĄNIMOUS COURT' "
                 "`play``the plays PLAYing SHOULDN'T CYCLE "
                 "cycling shouldn't uńąnimous or court's"},
                    [
                      {"name": 'Topic 1', "probability": 0.1647366},
                      {"name": 'Topic 2', "probability": 0.1885310},
                      {"name": 'Topic 3', "probability": 0.4879441},
                      {"name": 'Topic 4', "probability": 0.1587880}]

            ]
        ]

        for ex in examples:
            ex = dict(zip(headers, ex))
            show_method(self, self.bigml["method"], ex)
            lda_predict.i_make_a_prediction(
                self, ex["model"], ex["text"], ex["expected_distribution"])

    def test_scenario2(self):
        """
        Scenario 2: Successfully creating Topic Model from a dataset:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create topic model from a dataset
            And I wait until the topic model is ready less than <model_wait> secs
            And I update the topic model name to "<topic_model_name>"
            When I wait until the topic_model is ready less than <model_wait> secs
            Then the topic model name is "<topic_model_name>"
        """
        show_doc(self.test_scenario2)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "topic_model_name", "source_conf"]
        examples = [
            ['data/spam.csv', '100', '100', '100', 'my new topic model name', '{"fields": {"000001": {"optype": "text", "term_analysis": {"case_sensitive": true, "stem_words": true, "use_stopwords": false, "language": "en"}}}}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"])
            source_create.the_source_is_finished(self, example["source_wait"])
            source_create.i_update_source_with(self, example["source_conf"])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            topic_create.i_create_a_topic_model(self)
            topic_create.the_topic_model_is_finished_in_less_than(
                self, example["model_wait"])
            topic_create.i_update_topic_model_name(
                self, example["topic_model_name"])
            topic_create.the_topic_model_is_finished_in_less_than(
                self, example["model_wait"])
            topic_create.i_check_topic_model_name(
                self, example["topic_model_name"])



# -*- coding: utf-8 -*-
#pylint: disable=unbalanced-tuple-unpacking
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A class to deal with the fields of a resource.

This module helps to map between ids, names, and column_numbers in the
fields of source, dataset, or model. Also to validate your input data
for predictions or to list all the fields from a resource.

from bigml.api import BigML
from bigml.fields import Fields

api = BigML()

source = api.get_source("source/50a6bb94eabcb404d3000174")
fields = Fields(source['object']['fields'])

dataset = api.get_dataset("dataset/50a6bb96eabcb404cd000342")
fields = Fields(dataset['object']['fields'])

# Note that the fields in a model come one level deeper
model = api.get_model("model/50a6bbac035d0706db0008f8")
fields = Fields(model['object']['model']['fields'])

prediction = api.get_prediction("prediction/50a69688035d0706dd00044d")
fields =  Fields(prediction['object']['fields'])


"""
import sys
import json
import csv
import random


from bigml.util import invert_dictionary, python_map_type, find_locale
from bigml.util import DEFAULT_LOCALE
from bigml.api_handlers.resourcehandler import get_resource_type, get_fields
from bigml.constants import (
    SOURCE_PATH, DATASET_PATH, SUPERVISED_PATHS, FUSION_PATH,
    RESOURCES_WITH_FIELDS, DEFAULT_MISSING_TOKENS, REGIONS)
from bigml.io import UnicodeReader, UnicodeWriter

LIST_LIMIT = 10
REGIONS_ATTR = "labels"
SUMMARY_HEADERS = ["field column", "field ID", "field name", "field label",
                   "field description", "field type", "preferred",
                   "missing count", "errors", "contents summary",
                   "errors summary"]

UPDATABLE_HEADERS = {"field name": "name",
                     "field label": "label",
                     "field description": "description",
                     "field type": "optype",
                     "preferred": "preferred"}

ITEM_SINGULAR = {"categories": "category"}


def get_fields_structure(resource, errors=False):
    """Returns the field structure for a resource, its locale and
       missing_tokens

    """
    try:
        resource_type = get_resource_type(resource)
    except ValueError:
        raise ValueError("Unknown resource structure")
    field_errors = None
    resource = resource.get('object', resource)
    # locale and missing tokens
    if resource_type == SOURCE_PATH:
        resource_locale = resource['source_parser']['locale']
        missing_tokens = resource[
            'source_parser']['missing_tokens']
    else:
        resource_locale = resource.get('locale', DEFAULT_LOCALE)
        missing_tokens = resource.get('missing_tokens',
                                      DEFAULT_MISSING_TOKENS)

    fields = get_fields(resource)
    if resource_type in RESOURCES_WITH_FIELDS:
        # Check whether there's an objective id
        objective_column = None
        if resource_type == DATASET_PATH:
            objective_column = resource.get( \
                'objective_field', {}).get('id')
            if errors:
                field_errors = resource.get("status", {}).get("field_errors")
        elif resource_type in SUPERVISED_PATHS and \
                resource_type != FUSION_PATH:
            objective_id = resource.get( \
                'objective_fields', [None])[0]
            objective_column = fields.get( \
                objective_id, {}).get('column_number')
        result = fields, resource_locale, missing_tokens, objective_column
        if errors:
            result = result + (field_errors,)
        return result
    return (None, None, None, None, None) if errors else \
        (None, None, None, None)


def attribute_summary(attribute_value, item_type, limit=None):
    """Summarizes the information in fields attributes where content is
       written as an array of arrays like tag_cloud, items, etc.
    """
    if attribute_value is None:
        return None
    if item_type != REGIONS_ATTR:
        items = ["%s (%s)" % (item, instances) for
                 item, instances in attribute_value]
        items_length = len(items)
        if limit is None or limit > items_length:
            limit = items_length
        return "%s %s: %s" % (items_length, type_singular(item_type,
                                                          items_length == 1),
                              ", ".join(items[0: limit]))
    items = ["%s (%s)" % (attr.get("label"), attr.get("count")) for
             attr in attribute_value]
    items_length = len(items)
    if limit is None or limit > items_length:
        limit = items_length
    return "%s %s: %s" % (items_length, type_singular(item_type,
                                                      items_length == 1),
                          ", ".join(items[0: limit]))


def type_singular(item_type, singular=False):
    """Singularizes item types if needed

    """
    if singular:
        return ITEM_SINGULAR.get(item_type, item_type[:-1])
    return item_type


def numeric_example(numeric_summary):
    """Generates a random numeric example in the gaussian defined by
    mean and sigma in the numeric_summary

    """
    try:
        mean = numeric_summary.get("mean")
        sigma = numeric_summary.get("standard_deviation")
        minimum = numeric_summary.get("minimum")
        maximum = numeric_summary.get("maximum")
        value = -1
        while value < minimum or value > maximum:
            value = random.gauss(mean, sigma)
        return value
    except TypeError:
        return None


def sorted_headers(fields):
    """Listing the names of the fields as ordered in the original dataset.
    The `fields` parameter is a Fields object.
    """
    header_names = []
    header_ids = []
    for column in fields.fields_columns:
        header_names.append(fields.fields[
            fields.fields_by_column_number[column]]["name"])
        header_ids.append(fields.fields_by_column_number[column])

    return header_names, header_ids


def get_new_fields(output_fields):
    """Extracts the sexpr and names of the output fields in a dataset
    generated from a new_fields transformation.
    """
    new_fields = []
    for output_field in output_fields:
        sexp = output_field.get("generator")
        names = output_field.get("names")
        new_fields.append({"field": sexp, "names": names})
    return new_fields


class Fields():
    """A class to deal with BigML auto-generated ids.

    """
    def __init__(self, resource_or_fields, missing_tokens=None,
                 data_locale=None, verbose=False,
                 objective_field=None, objective_field_present=False,
                 include=None, errors=None):

        # The constructor can be instantiated with resources or a fields
        # structure. The structure is checked and fields structure is returned
        # if a resource type is matched.
        try:
            self.resource_type = get_resource_type(resource_or_fields)
            resource_info = get_fields_structure(resource_or_fields, True)
            (self.fields,
             resource_locale,
             resource_missing_tokens,
             objective_column,
             resource_errors) = resource_info
            if data_locale is None:
                data_locale = resource_locale
            if missing_tokens is None:
                if resource_missing_tokens:
                    missing_tokens = resource_missing_tokens
            if errors is None:
                errors = resource_errors
        except ValueError:
            # If the resource structure is not in the expected set, fields
            # structure is assumed
            self.fields = resource_or_fields
            if data_locale is None:
                data_locale = DEFAULT_LOCALE
            if missing_tokens is None:
                missing_tokens = DEFAULT_MISSING_TOKENS
            objective_column = None
        if self.fields is None:
            raise ValueError("No fields structure was found.")
        self.fields_by_name = invert_dictionary(self.fields, 'name')
        self.fields_by_column_number = invert_dictionary(self.fields,
                                                         'column_number')
        find_locale(data_locale, verbose)
        self.missing_tokens = missing_tokens
        self.fields_columns = sorted(self.fields_by_column_number.keys())
        # Ids of the fields to be included
        self.filtered_fields = (list(self.fields.keys()) if include is None
                                else include)
        # To be updated in update_objective_field
        self.row_ids = None
        self.headers = None
        self.objective_field = None
        self.objective_field_present = None
        self.filtered_indexes = None
        self.field_errors = errors
        # if the objective field is not set by the user
        # use the one extracted from the resource info
        if objective_field is None and objective_column is not None:
            objective_field = objective_column
            objective_field_present = True
        if self.fields:
            # empty composite sources will not have an objective field
            self.update_objective_field(objective_field,
                                        objective_field_present)

    def update_objective_field(self, objective_field, objective_field_present,
                               headers=None):
        """Updates objective_field and headers info

            Permits to update the objective_field, objective_field_present and
            headers info from the constructor and also in a per row basis.
        """
        # If no objective field, select the last column, else store its column
        if objective_field is None:
            self.objective_field = self.fields_columns[-1]
        elif isinstance(objective_field, str):
            try:
                self.objective_field = self.field_column_number( \
                    objective_field)
            except KeyError:
                # if the name of the objective field is not found, use the last
                # field as objective
                self.objective_field = self.fields_columns[-1]
        else:
            self.objective_field = objective_field

        # If present, remove the objective field from the included fields
        objective_id = self.field_id(self.objective_field)
        if objective_id in self.filtered_fields:
            del self.filtered_fields[self.filtered_fields.index(objective_id)]

        self.objective_field_present = objective_field_present
        if headers is None:
            # The row is supposed to contain the fields sorted by column number
            self.row_ids = [item[0] for item in
                            sorted(list(self.fields.items()),
                                   key=lambda x: x[1]['column_number'])
                            if objective_field_present or
                            item[1]['column_number'] != self.objective_field]
            self.headers = self.row_ids
        else:
            # The row is supposed to contain the fields as sorted in headers
            self.row_ids = [self.field_id(header) for header in headers]
            self.headers = headers
        # Mapping each included field to its correspondent index in the row.
        # The result is stored in filtered_indexes.
        self.filtered_indexes = []
        for field in self.filtered_fields:
            try:
                index = self.row_ids.index(field)
                self.filtered_indexes.append(index)
            except ValueError:
                continue

    def field_id(self, key):
        """Returns a field id.

        """

        if isinstance(key, str):
            try:
                f_id = self.fields_by_name[key]
            except KeyError:
                raise ValueError("Error: field name '%s' does not exist" % key)
            return f_id
        if isinstance(key, int):
            try:
                f_id = self.fields_by_column_number[key]
            except KeyError:
                raise ValueError("Error: field column number '%s' does not"
                                 " exist" % key)
            return f_id
        return None

    def field_name(self, key):
        """Returns a field name.

        """
        if isinstance(key, str):
            try:
                name = self.fields[key]['name']
            except KeyError:
                raise ValueError("Error: field id '%s' does not exist" % key)
            return name
        if isinstance(key, int):
            try:
                name = self.fields[self.fields_by_column_number[key]]['name']
            except KeyError:
                raise ValueError("Error: field column number '%s' does not"
                                 " exist" % key)
            return name
        return None

    def field_column_number(self, key):
        """Returns a field column number.

        """
        try:
            return self.fields[key]['column_number']
        except KeyError:
            return self.fields[self.fields_by_name[key]]['column_number']

    def len(self):
        """Returns the number of fields.

        """
        return len(self.fields)

    def pair(self, row, headers=None,
             objective_field=None, objective_field_present=None):
        """Pairs a list of values with their respective field ids.

            objective_field is the column_number of the objective field.

           `objective_field_present` must be True is the objective_field column
           is present in the row.

        """
        # Try to get objective field form Fields or use the last column
        if objective_field is None:
            if self.objective_field is None:
                objective_field = self.fields_columns[-1]
            else:
                objective_field = self.objective_field
        # If objective fields is a name or an id, retrive column number
        if isinstance(objective_field, str):
            objective_field = self.field_column_number(objective_field)

        # Try to guess if objective field is in the data by using headers or
        # comparing the row length to the number of fields
        if objective_field_present is None:
            if headers:
                objective_field_present = (self.field_name(objective_field) in
                                           headers)
            else:
                objective_field_present = len(row) == self.len()

        # If objective field, its presence or headers have changed, update
        if (objective_field != self.objective_field or
                objective_field_present != self.objective_field_present or
                (headers is not None and headers != self.headers)):
            self.update_objective_field(objective_field,
                                        objective_field_present, headers)

        row = [self.normalize(info) for info in row]
        return self.to_input_data(row)

    def list_fields(self, out=sys.stdout):
        """Lists a description of the fields.

        """
        for field in [(val['name'], val['optype'], val['column_number'])
                      for _, val in sorted(list(self.fields.items()),
                                           key=lambda k:
                                           k[1]['column_number'])]:
            out.write('[%-32s: %-16s: %-8s]\n' % (field[0],
                                                  field[1], field[2]))
            out.flush()

    def preferred_fields(self):
        """Returns fields where attribute preferred is set to True or where
           it isn't set at all.

        """
        return {key: field for key, field in self.fields.items()
                if ('preferred' not in field) or field['preferred']}

    def validate_input_data(self, input_data, out=sys.stdout):
        """Validates whether types for input data match types in the
        fields definition.

        """
        if isinstance(input_data, dict):
            for name in input_data:
                if name in self.fields_by_name:
                    out.write('[%-32s: %-16s: %-16s: ' %
                              (name, type(input_data[name]),
                               self.fields[self.fields_by_name[name]]
                               ['optype']))
                    if (type(input_data[name]) in python_map_type(self.fields[
                            self.fields_by_name[name]]['optype'])):
                        out.write('OK\n')
                    else:
                        out.write('WRONG\n')
                else:
                    out.write("Field '%s' does not exist\n" % name)
        else:
            out.write("Input data must be a dictionary")

    def normalize(self, value):
        """Transforms to unicode and cleans missing tokens

        """
        if not isinstance(value, str):
            value = str(value, "utf-8")
        return None if value in self.missing_tokens else value

    def to_input_data(self, row):
        """Builds dict with field, value info only for the included headers

        """
        pair = []
        for index in self.filtered_indexes:
            pair.append((self.headers[index], row[index]))
        return dict(pair)

    def missing_counts(self):
        """Returns the ids for the fields that contain missing values

        """
        summaries = [(field_id, field.get('summary', {}))
                     for field_id, field in list(self.fields.items())]
        if len(summaries) == 0:
            raise ValueError("The structure has not enough information "
                             "to extract the fields containing missing values."
                             "Only datasets and models have such information. "
                             "You could retry the get remote call "
                             " with 'limit=-1' as query string.")

        return {field_id: summary.get('missing_count', 0)
                for field_id, summary in summaries
                if summary.get('missing_count', 0) > 0}

    def stats(self, field_name):
        """Returns the summary information for the field

        """
        field_id = self.field_id(field_name)
        summary = self.fields[field_id].get('summary', {})
        return summary

    def summary_csv(self, filename=None):
        """Summary of the contents of the fields

        """

        summary = []
        writer = None
        if filename is not None:
            writer = UnicodeWriter(filename,
                                   quoting=csv.QUOTE_NONNUMERIC).open_writer()
            writer.writerow(SUMMARY_HEADERS)
        else:
            summary.append(SUMMARY_HEADERS)

        for field_column in self.fields_columns:
            field_id = self.field_id(field_column)
            field = self.fields.get(field_id)
            field_summary = []
            field_summary.append(field.get('column_number'))
            field_summary.append(field_id)
            field_summary.append(field.get('name'))
            field_summary.append(field.get('label'))
            field_summary.append(field.get('description'))
            field_summary.append(field.get('optype'))
            field_summary_value = field.get('summary', {})

            if not field_summary_value:
                field_summary.append("") # no preferred info
                field_summary.append("") # no missing info
                field_summary.append("") # no error info
                field_summary.append("") # no content summary
                field_summary.append("") # no error summary
            else:
                field_summary.append(json.dumps(field.get('preferred')))
                field_summary.append(field_summary_value.get("missing_count"))
                if self.field_errors and field_id in list(self.field_errors.keys()):
                    errors = self.field_errors.get(field_id)
                    field_summary.append(errors.get("total"))
                else:
                    field_summary.append("0")
                if field['optype'] == 'numeric':
                    field_summary.append("[%s, %s], mean: %s" % \
                        (field_summary_value.get("minimum"),
                         field_summary_value.get("maximum"),
                         field_summary_value.get("mean")))
                elif field['optype'] == 'categorical':
                    categories = field_summary_value.get("categories")
                    field_summary.append( \
                        attribute_summary(categories, "categorìes",
                                          limit=LIST_LIMIT))
                elif field['optype'] == REGIONS:
                    labels_info = field_summary_value.get("labels")
                    field_summary.append( \
                        attribute_summary(labels_info, "labels",
                                          limit=LIST_LIMIT))
                elif field['optype'] == "text":
                    terms = field_summary_value.get("tag_cloud")
                    field_summary.append( \
                        attribute_summary(terms, "terms",
                                          limit=LIST_LIMIT))
                elif field['optype'] == "items":
                    items = field_summary_value.get("items")
                    field_summary.append( \
                        attribute_summary(items, "items", limit=LIST_LIMIT))
                else:
                    field_summary.append("")
                if self.field_errors and field_id in list(self.field_errors.keys()):
                    field_summary.append( \
                        attribute_summary(errors.get("sample"), "errors",
                                          limit=None))
                else:
                    field_summary.append("")
            if writer:
                writer.writerow(field_summary)
            else:
                summary.append(field_summary)
        if writer is None:
            return summary
        writer.close_writer()
        return filename

    def new_fields_structure(self, csv_attributes_file=None,
                             attributes=None, out_file=None):
        """Builds the field structure needed to update a fields dictionary
        in a BigML resource.

        :param csv_attributes_file: (string) Path to a CSV file like the one
                                             generated by summary_csv.
        :param attributes: (list) list of rows containing the
                                  attributes information ordered
                                  as in the summary_csv output.
        :param out_file: (string) Path to a JSON file that will be used
                                  to store the new fields structure. If None,
                                  the output is returned as a dict.
        """
        if csv_attributes_file is not None:
            reader = UnicodeReader(csv_attributes_file).open_reader()
            attributes = list(reader)
        new_fields_structure = {}
        if "field ID" in attributes[0] or "field column" in attributes[0]:
            # headers are used
            for index in range(1, len(attributes)):
                new_attributes = dict(list(zip(attributes[0],
                                               attributes[index])))
                if new_attributes.get("field ID"):
                    field_id = new_attributes.get("field ID")
                    if not field_id in list(self.fields.keys()):
                        raise ValueError("Field ID %s not found"
                                         " in this resource" % field_id)
                    del new_attributes["field ID"]
                else:
                    try:
                        field_column = int(new_attributes.get("field column"))
                    except TypeError:
                        raise ValueError(
                            "Field column %s not found"
                            " in this resource" % new_attributes.get(
                            "field_column"))
                    if not field_column in self.fields_columns:
                        raise ValueError("Field column %s not found"
                                         " in this resource" % field_column)
                    field_id = self.field_id(field_column)
                    del new_attributes["field column"]
                new_attributes_headers = list(new_attributes.keys())
                for attribute in new_attributes_headers:
                    if not attribute in list(UPDATABLE_HEADERS.keys()):
                        del new_attributes[attribute]
                    else:
                        new_attributes[UPDATABLE_HEADERS[attribute]] = \
                            new_attributes[attribute]
                        if attribute != UPDATABLE_HEADERS[attribute]:
                            del new_attributes[attribute]
                if "preferred" in new_attributes:
                    new_attributes['preferred'] = json.loads( \
                        new_attributes['preferred'])
                new_fields_structure[field_id] = new_attributes
        else:
            # assume the order given in the summary_csv method
            first_attribute = attributes[0][0]
            first_column_is_id = False
            try:
                field_id = self.field_id(int(first_attribute))
            except ValueError:
                field_id = first_attribute
                first_column_is_id = True
            if not field_id in self.fields:
                raise ValueError("The first column should contain either the"
                                 " column or ID of the fields. Failed to find"
                                 " %s as either of them." % field_id)
            headers = SUMMARY_HEADERS[2: 7]
            headers = [UPDATABLE_HEADERS[header] for header in headers]
            try:
                for field_attributes in attributes:
                    if field_attributes[6] is not None:
                        field_attributes[6] = json.loads(field_attributes[6])
                    field_id = field_attributes[0] if first_column_is_id else \
                        self.field_id(int(field_attributes[0]))
                    new_fields_structure[field_id] = \
                        dict(list(zip(headers, field_attributes[1: 6])))
            except ValueError:
                raise ValueError("The first column should contain either the"
                                 " column or ID of the fields. Failed to find"
                                 " %s as either of them." % field_id)
        new_fields_structure = {"fields": new_fields_structure}
        if out_file is None:
            return new_fields_structure
        try:
            with open(out_file, "w") as out:
                json.dump(new_fields_structure, out)
        except IOError:
            raise IOError("Failed writing the fields structure file in"
                          " %s- Please, check your arguments." %
                          out_file)
        return out_file

    def training_data_example(self, missings=False):
        """Generates an example of training data based on the contents of the
        summaries of every field

        If missings is set to true, missing values are allowed

        """
        training_data = {}
        for _, field in list(self.fields.items()):
            if field.get("summary") is not None:
                value = None
                optype = field.get("optype")
                if optype == "numeric":
                    if missings and random.randint(0, 5) > 3:
                        value = None
                    else:
                        value = numeric_example(field["summary"])
                if optype == "categorical":
                    if missings and random.randint(0, 5) > 3:
                        value = None
                    else:
                        categories = [cat[0] for cat in field["summary"]["categories"]]
                        weights = [cat[1] for cat in field["summary"]["categories"]]
                        value = random.choices(categories, weights)[0]
                if optype == "text":
                    if missings and random.randint(0, 5) > 3:
                        value = None
                    else:
                        text_number = len(field["summary"]["tag_cloud"])
                        index = random.randint(0, text_number - 1)
                        value = field["summary"]["tag_cloud"][index][0]
                if optype == "items":
                    if missings and random.randint(0, 5) > 3:
                        value = None
                    else:
                        items_number = len(field["summary"]["items"])
                        index = random.randint(0, items_number - 1)
                        value = field["summary"]["items"][index][0]
                if optype == REGIONS:
                    if missings and random.randint(0, 5) > 3:
                        value = None
                    else:
                        labels_number = len(field["summary"]["labels"])
                        index = random.randint(0, labels_number - 1)
                        field_summary = field["summary"]["labels"][index]
                        label = field_summary["label"]
                        xmin = numeric_example(field_summary["xmin"])
                        xmax = numeric_example(field_summary["xmax"])
                        ymin = numeric_example(field_summary["ymin"])
                        ymax = numeric_example(field_summary["ymax"])
                        #pylint: disable=locally-disabled,too-many-boolean-expressions
                        if None in [xmin, xmax, ymin, ymax] or xmax < xmin or \
                                ymax < ymin or xmin < 0 or xmax < 0 or \
                                ymin < 0 or ymax < 0:
                            value = []
                        else:
                            value = [[label, xmin, xmax, ymin, ymax]]

                if value is not None:
                    training_data.update({field["name"]: value})
        return training_data

    def filter_fields_update(self, update_body):
        """Filters the updatable attributes according to the type of resource

        """
        fields_info = update_body.get("fields")
        if self.resource_type and fields_info is not None:
            if self.resource_type == "dataset":
                for _, field in list(fields_info.items()):
                    if field.get("optype") is not None:
                        del field["optype"]
            elif self.resource_type == "source":
                for _, field in list(fields_info.items()):
                    if field.get("preferred") is not None:
                        del field["preferred"]
            update_body["fields"] = fields_info
        return update_body



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Testing local prediction

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import compare_predictions_steps as prediction_compare
from . import create_ensemble_steps as ensemble_create
from . import create_prediction_steps as prediction_create


class TestLocalPrediction:
    """Testing local predictions """

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario 1: Successfully creating a prediction from a local model in a json file:
            Given I create a local model from a "<model>" file
            When I create a local prediction for "<data_input>" with confidence
            Then the local prediction is "<prediction>"
            And the local prediction's confidence is "<confidence>"
        """
        show_doc(self.test_scenario1)
        headers = ["file_path", "input_data", "prediction", "confidence"]
        examples = [
            ['data/iris_model.json', '{"petal length": 0.5}', 'Iris-setosa',
             '0.90594'],
            ['data/iris_model.json', '{"petal length": "0.5"}', 'Iris-setosa',
             '0.90594']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            prediction_compare.i_create_a_local_model_from_file(
                self, example["file_path"])
            prediction_compare.i_create_a_local_prediction_with_confidence(
                self, example["input_data"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])
            prediction_compare.the_local_prediction_confidence_is(
                self, example["confidence"])

    def test_scenario2(self):
        """
        Scenario 2: Successfully creating a prediction from a local model in a json file:
            Given I create a local model using SupervisedModel from a "<model>" file
            When I create a local prediction for "<data_input>" with confidence
            Then the local prediction is "<prediction>"
            And the local prediction's confidence is "<confidence>"
        """
        show_doc(self.test_scenario2)
        headers = ["file_path", "input_data", "prediction", "confidence"]
        examples = [
            ['data/iris_model.json', '{"petal length": 0.5}', 'Iris-setosa',
             '0.90594'],
            ['data/iris_model.json', '{"petal length": "0.5"}', 'Iris-setosa',
             '0.90594']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            prediction_compare.i_create_a_local_supervised_model_from_file(
                self, example["file_path"])
            prediction_compare.i_create_a_local_prediction_with_confidence(
                self, example["input_data"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])
            prediction_compare.the_local_prediction_confidence_is(
                self, example["confidence"])


    def test_scenario3(self):
        """
        Scenario 3: Successfully creating a local prediction from an Ensemble created from file storage:
            Given I create a local Ensemble from path "<path>"
            When I create a local ensemble prediction with confidence for "<data_input>"
            Then the local prediction is "<prediction>"
            And the local prediction's confidence is "<confidence>"
            And the local probabilities are "<probabilities>"
        """
        show_doc(self.test_scenario3)
        headers = ["file_path", "input_data", "prediction", "confidence",
                   "probabilities"]
        examples = [
            ['bigml/tests/my_no_root_ensemble/ensemble.json',
             '{"petal width": 0.5}', 'Iris-setosa', '0.3533',
             '["0.3533", "0.31", "0.33666"]' ]]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            ensemble_create.create_local_ensemble(
                self, path=example["file_path"])
            prediction_create.create_local_ensemble_prediction_probabilities(
                self, example["input_data"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])
            prediction_compare.the_local_prediction_confidence_is(
                self, example["confidence"])
            prediction_compare.the_local_probabilities_are(
                self, example["probabilities"])

    def test_scenario4(self):
        """
        Scenario 4: Successfully creating a local prediction from an Ensemble created from file storage:
            Given I create a local SupervisedModel from path "<path>"
            When I create a local ensemble prediction with confidence for "<data_input>"
            Then the local prediction is "<prediction>"
            And the local prediction's confidence is "<confidence>"
            And the local probabilities are "<probabilities>"
        """
        show_doc(self.test_scenario4)
        headers = ["file_path", "input_data", "prediction", "confidence",
                   "probabilities"]
        examples = [
            ['bigml/tests/my_no_root_ensemble/ensemble.json',
             '{"petal width": 0.5}', 'Iris-setosa', '0.3533',
             '["0.3533", "0.31", "0.33666"]' ]]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            prediction_compare.i_create_a_local_supervised_model_from_file(
                self, example["file_path"])
            prediction_compare.i_create_a_local_prediction_with_confidence(
                self, example["input_data"])
            prediction_compare.i_create_local_probabilities(
                self, example["input_data"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])
            prediction_compare.the_local_prediction_confidence_is(
                self, example["confidence"])
            prediction_compare.the_local_probabilities_are(
                self, example["probabilities"])

    def test_scenario5(self):
        """
        Scenario 5: Successfully creating a prediction from a local images deepnet in a json file:
            Given I create a local deepnet from a "<deepnet>" file
            When I create a local prediction for "<data_input>"
            Then the local prediction is "<prediction>"
        """
        show_doc(self.test_scenario5)
        headers = ["file_path", "input_data", "operation_settings",
                   "prediction"]
        examples = [
            ['data/imgs_deepnet.zip', "data/images/cats/pexels-pixabay-33358.jpg",
             {"region_score_threshold": 0.7},
             '{"prediction": [{"box": [0.68164, 0.30469, 0.79688, 0.36979], "label": "eye", "score": 0.79633}, {"box": [0.38086, 0.27865, 0.50391, 0.36068], "label": "eye", "score": 0.74563}]}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            prediction_compare.i_create_a_local_deepnet_from_zip_file(
                self, example["file_path"],
                operation_settings=example["operation_settings"])
            prediction_compare.i_create_a_local_regions_prediction(
                self, example["input_data"])
            prediction_compare.the_local_regions_prediction_is(
                self, example["prediction"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Testing MultiVote predictions

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import compute_multivote_prediction_steps as multivote_prediction


class TestMultiVotePrediction:
    """Testing MultiVote methods"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario 1: Successfully computing predictions combinations:
            Given I create a MultiVote for the set of predictions in file <predictions>
            When I compute the prediction with confidence using method "<method>"
            And I compute the prediction without confidence using method "<method>"
            Then the combined prediction is "<prediction>"
            And the combined prediction without confidence is "<prediction>"
            And the confidence for the combined prediction is <confidence>
        """
        show_doc(self.test_scenario1)
        headers = ["predictions_file", "method", "prediction", "confidence"]
        examples = [
            ['data/predictions_c.json', '0', 'a', '0.45047'],
            ['data/predictions_c.json', '1', 'a', '0.55202'],
            ['data/predictions_c.json', '2', 'a', '0.40363'],
            ['data/predictions_r.json', '0', '1.55555556667', '0.40008'],
            ['data/predictions_r.json', '1', '1.59376845074', '0.24837'],
            ['data/predictions_r.json', '2', '1.55555556667', '0.40008']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            multivote_prediction.i_create_a_multivote(
                self, example["predictions_file"])
            multivote_prediction.compute_prediction(
            self, example["method"])
            multivote_prediction.compute_prediction_no_confidence(
                self, example["method"])
            multivote_prediction.check_combined_prediction(
                self, example["prediction"])
            multivote_prediction.check_combined_prediction_no_confidence(
                self, example["prediction"])
            multivote_prediction.check_combined_confidence(
                self, example["confidence"])
